# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: app.py 4 2008-02-26 05:04:22Z albrecht.andi $

"""Application object"""

import bonobo

import sys

from cf import release
from config import Config # IGNORE:E0611
from datasources import DatasourceManager
from plugins.core import PluginManager
from ui.prefs import PreferencesDialog
from userdb import UserDB

import logging
log = logging.getLogger("APP")

class CFApplication(bonobo.Application):
    """Main application object
    
    :Attributes:
        config: preferences
        userdb: uer database
        plugins: pugin manager
        datasources: datasource manager
    
    """
    
    def __init__(self):
        bonobo.Application.__init__(self, release.name)
        self.__shutdown_tasks = []
        self.config = Config(self)
        self.userdb = UserDB(self)
        self.plugins = PluginManager(self)
        self.datasources = DatasourceManager(self)
        
    def preferences_show(self):
        """Displays the preferences dialog"""
        dlg = PreferencesDialog(self)
        dlg.run() # IGNORE:E1101 - Generic method
        dlg.destroy() # IGNORE:E1101 - Generic method
        
    def register_shutdown_task(self, func, msg, *args, **kwargs):
        """Registers a new shutdown task.
        
        :param func: a callable
        :param msg: a human-readable message explaining what this task does
        :param args, kwargs: arguments and kw arguments for this task
        """
        self.__shutdown_tasks.append((func, msg, args, kwargs))
        
    def shutdown(self):
        """Execute all shutdown tasks and quit application"""
        while self.__shutdown_tasks:
            func, msg, args, kwargs = self.__shutdown_tasks.pop()
            log.info(msg)
            try:
                func(*args, **kwargs)
            except:
                log.error("Task failed: %s" % str(sys.exc_info()[1])) 
        


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: prefs.py 4 2008-02-26 05:04:22Z albrecht.andi $

"""Preferences"""

import gtk
import gobject
import gnome
import gconf

from gettext import gettext as _

import cf
from cf.ui import GladeWidget

class PreferencesDialog(GladeWidget):
    
    def __init__(self, app):
        GladeWidget.__init__(self, app, "crunchyfrog", "preferences_dialog")
        self.refresh()
        
    def _setup_widget(self):
        self._setup_plugins()
        self._setup_editor()
        # Fix secondary button
        btn = self.xml.get_widget("btn_help")
        box = self.xml.get_widget("dialog-action_area1")
        box.set_child_secondary(btn, True)
        
    def _setup_editor(self):
        model = gtk.ListStore(int, str, gtk.gdk.Pixbuf)
        iconview = self.xml.get_widget("editor_iconview")
        iconview.set_model(model)
        iconview.set_text_column(1)
        iconview.set_pixbuf_column(2)
        it = gtk.icon_theme_get_default()
        def get_pb(stock):
            return it.load_icon(stock, 36, gtk.ICON_LOOKUP_FORCE_SVG)
        model.append([0, _(u"General"), get_pb("gtk-execute")])
        model.append([1, _(u"View"), get_pb("gtk-justify-left")])
        model.append([2, _(u"Editor"), get_pb("gtk-edit")])
        model.append([3, _(u"Fonts & Colors"), get_pb("gtk-font")])
        iconview.connect("selection-changed", self.on_editor_selection_changed)
        from cf.ui.editor import USE_GTKSOURCEVIEW2
        if not USE_GTKSOURCEVIEW2:
            self.xml.get_widget("editor_schemes_box").set_sensitive(False)
            self.xml.get_widget("editor_schemes_box").hide()
            # FIXME: Is there really no way to set right margin position on SourceView?
            self.xml.get_widget("editor_right_margin_position_box").set_sensitive(False)
            self.xml.get_widget("editor_right_margin_position_box").hide()
        else:
            schemes = self.xml.get_widget("editor_schemes")
            model = gtk.ListStore(str, str)
            model.set_sort_column_id(1, gtk.SORT_ASCENDING)
            schemes.set_model(model)
            col = gtk.TreeViewColumn("", gtk.CellRendererText(), markup=1)
            schemes.append_column(col)
            import gtksourceview2
            sm = gtksourceview2.style_scheme_manager_get_default()
            selected = None
            for id in sm.get_scheme_ids():
                scheme = sm.get_scheme(id)
                lbl = "<b>%s</b>" % scheme.get_name()
                if scheme.get_description():
                    lbl += "\n"+scheme.get_description()
                iter = model.append(None)
                model.set(iter, 0, id, 1, lbl)
                if id == self.app.config.get("editor.scheme"):
                    selected = iter
            sel = schemes.get_selection()
            sel.select_iter(selected)
            sel.connect("changed", self.on_editor_option_changed)
            sel.set_data("config_option", "editor.scheme")
                
        
    def _setup_plugins(self):
        """Set up the plugins view"""
        self.plugin_model = gtk.TreeStore(gobject.TYPE_PYOBJECT, # 0 Plugin class
                                          bool, # 1 active
                                          str, # 2 label
                                          gtk.gdk.Pixbuf, # 3 icon
                                          bool, # 4 active visible
                                          )
        self.plugin_model.set_sort_column_id(2, gtk.SORT_ASCENDING)
        self.plugin_list = self.xml.get_widget("plugin_list")
        self.plugin_list.set_model(self.plugin_model)
        # label
        col = gtk.TreeViewColumn()
        renderer = gtk.CellRendererPixbuf()
        col.pack_start(renderer, expand=False)
        col.add_attribute(renderer, 'pixbuf', 3)
        renderer = gtk.CellRendererText()
        col.pack_start(renderer, expand=True)
        col.add_attribute(renderer, 'markup', 2)
        self.plugin_list.append_column(col)
        # active
        renderer = gtk.CellRendererToggle()
        renderer.connect("toggled", self.on_plugin_active_toggled)
        col = gtk.TreeViewColumn("", renderer, active=1, visible=4)
        self.plugin_list.append_column(col)
        
    def _setup_connections(self):
        self.app.plugins.connect("plugin-added", self.on_plugin_added)
        self.app.plugins.connect("plugin-removed", self.on_plugin_removed)
        
    def on_editor_open_in_window_toggled(self, toggle):
        self.app.config.set("editor.open_in_window", toggle.get_active())
        
    def on_editor_option_changed(self, widget, *args):
        option = widget.get_data("config_option")
        conf = self.app.config
        if isinstance(widget, gtk.CheckButton):
            conf.set(option, widget.get_active())
        elif isinstance(widget, gtk.SpinButton):
            conf.set(option, widget.get_value_as_int())
        elif isinstance(widget, gtk.FontButton):
            conf.set(option, widget.get_font_name())
        elif isinstance(widget, gtk.TreeSelection):
            model, iter = widget.get_selected()
            conf.set(option, model.get_value(iter, 0))
        if option == "editor.wrap_text":
            self.xml.get_widget("editor_wrap_split").set_sensitive(widget.get_active())
        if option == "editor.right_margin":
            self.xml.get_widget("editor_right_margin_position_box").set_sensitive(widget.get_active())
        if option == "editor.default_font":
            self.xml.get_widget("editor_font_box").set_sensitive(not widget.get_active())
        
    def on_editor_selection_changed(self, iconview):
        model = iconview.get_model()
        for path in iconview.get_selected_items():
            iter = model.get_iter(path)
            nb = self.xml.get_widget("editor_notebook")
            nb.set_current_page(model.get_value(iter, 0))
    
    def on_plugin_active_toggled(self, renderer, path):
        iter = self.plugin_model.get_iter(path)
        self.plugin_model.set_value(iter, 1, not renderer.get_active())
        plugin = self.plugin_model.get_value(iter, 0)
        gobject.idle_add(self.app.plugins.set_active, plugin, not renderer.get_active())
        
    def on_plugin_added(self, manager, plugin):
        iter = self.plugin_model.get_iter_first()
        it = gtk.icon_theme_get_default()
        while iter:
            if self.plugin_model.get_value(iter, 0) == plugin._entry_point_group:
                break
            iter = self.plugin_model.iter_next(iter)
        if iter:
            lbl = '<b>%s</b>' % plugin.name or _(u"Unknown")
            if plugin.description:
                lbl += "\n"+plugin.description
            if plugin.icon:
                ico = it.load_icon(plugin.icon, gtk.ICON_SIZE_LARGE_TOOLBAR, gtk.ICON_LOOKUP_FORCE_SVG)
            else:
                ico = None
            citer = self.plugin_model.append(iter)
            self.plugin_model.set(citer,
                      0, plugin,
                      1, self.app.plugins.is_active(plugin),
                      2, lbl,
                      3, ico,
                      4, True)
            
    def on_plugin_removed(self, manager, plugin):
        iter = self.plugin_model.get_iter_first()
        while iter:
            if self.plugin_model.get_value(iter, 0) == plugin._entry_point_group:
                citer = self.plugin_model.iter_children(iter)
                while citer:
                    if self.plugin_model.get_value(citer, 0) == plugin:
                        self.plugin_model.remove(citer)
                        return
                    citer = self.plugin_model.iter_next(citer)
            iter = self.plugin_model.iter_next(iter)
                
    def on_plugin_folder_show(self, *args):
        gnome.url_show(cf.USER_PLUGIN_URI)
        
    def refresh(self):
        self.refresh_editor()
        self.refresh_plugins()
        
    def refresh_editor(self):
        config = self.app.config
        gw = self.xml.get_widget
        gw("editor_open_separate").set_active(config.get("editor.open_in_window"))
        gw("editor_wrap_text").set_data("config_option", "editor.wrap_text")
        gw("editor_wrap_text").set_active(config.get("editor.wrap_text"))
        gw("editor_wrap_split").set_data("config_option", "editor.wrap_split")
        gw("editor_wrap_split").set_active(config.get("editor.wrap_split"))
        gw("editor_wrap_split").set_sensitive(gw("editor_wrap_text").get_active())
        gw("editor_display_line_numbers").set_data("config_option", "editor.display_line_numbers")
        gw("editor_display_line_numbers").set_active(config.get("editor.display_line_numbers"))
        gw("editor_highlight_current_line").set_data("config_option", "editor.highlight_current_line")
        gw("editor_highlight_current_line").set_active(config.get("editor.highlight_current_line"))
        gw("editor_right_margin").set_data("config_option", "editor.right_margin")
        gw("editor_right_margin").set_active(config.get("editor.right_margin"))
        gw("editor_right_margin_position").set_data("config_option", "editor.right_margin_position")
        gw("editor_right_margin_position").set_value(config.get("editor.right_margin_position"))
        gw("editor_right_margin_position_box").set_sensitive(config.get("editor.right_margin"))
        gw("editor_bracket_matching").set_data("config_option", "editor.bracket_matching")
        gw("editor_bracket_matching").set_active(config.get("editor.bracket_matching"))
        gw("editor_tabs_width").set_data("config_option", "editor.tabs_width")
        gw("editor_tabs_width").set_value(config.get("editor.tabs_width"))
        gw("editor_insert_spaces").set_data("config_option", "editor.insert_spaces")
        gw("editor_insert_spaces").set_active(config.get("editor.insert_spaces"))
        gw("editor_auto_indent").set_data("config_option", "editor.auto_indent")
        gw("editor_auto_indent").set_active(config.get("editor.auto_indent"))
        gw("editor_default_font").set_data("config_option", "editor.default_font")
        gw("editor_default_font").set_active(config.get("editor.default_font"))
        client = gconf.client_get_default()
        default_font = client.get_string("/desktop/gnome/interface/monospace_font_name")
        gw("editor_default_font").set_label(gw("editor_default_font").get_label() % default_font)
        gw("editor_font_box").set_sensitive(not config.get("editor.default_font"))
        gw("editor_font").set_data("config_option", "editor.font")
        gw("editor_font").set_font_name(config.get("editor.font"))
        
    def refresh_plugins(self):
        # Plugins
        model = self.plugin_model
        model.clear()
        it = gtk.icon_theme_get_default()
        for key, value in self.app.plugins.entry_points.items():
            iter = model.append(None)
            model.set(iter,
                      0, key, 
                      1, False,
                      2, '<b>%s</b>' % value[0],
                      3, None,
                      4, False)
            for plugin in self.app.plugins.get_plugins(key):
                lbl = '<b>%s</b>' % plugin.name or _(u"Unknown")
                if plugin.description:
                    lbl += "\n"+plugin.description
                if plugin.icon:
                    ico = it.load_icon(plugin.icon, gtk.ICON_SIZE_LARGE_TOOLBAR, gtk.ICON_LOOKUP_FORCE_SVG)
                else:
                    ico = None
                citer = model.append(iter)
                model.set(citer,
                          0, plugin,
                          1, self.app.plugins.is_active(plugin),
                          2, lbl,
                          3, ico,
                          4, True)


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: oracle.py 16 2008-02-26 14:52:30Z albrecht.andi $

"""Oracle backend"""

from cf.backends import DBConnectError
from cf.backends.dbapi2helper import DbAPI2Connection, DbAPI2Cursor
from cf.backends.schema import *
from cf.datasources import DatasourceInfo
from cf.plugins.core import DBBackendPlugin

import time

from gettext import gettext as _

import logging
log = logging.getLogger("ORACLE")

class OracleBackend(DBBackendPlugin):
    name = _(u"Oracle Plugin")
    description = _(u"Provides access to Oracle databases")
    
    def __init__(self, app):
        DBBackendPlugin.__init__(self, app)
        log.info("Activating Oracle backend")
        self.schema = OracleSchema()
        
    def _get_conn_opts_from_opts(self, opts):
        conn_opts = dict()
        if opts["database"]: conn_opts["dsn"] = opts["database"]
        if opts["user"]: conn_opts["user"] = opts["user"]
        if opts["password"]: conn_opts["password"] = opts["password"]
        return conn_opts
        
    def shutdown(self):
        log.info("Shutting down Oracle backend")
    
    @classmethod
    def get_datasource_options_widgets(cls, data_widgets, initial_data=None):
        return data_widgets, ["database", "user", "password"]
    
    @classmethod
    def get_label(cls, datasource_info):
        s = "%s@%s" % (datasource_info.options.get("user", None) or "??",
                       datasource_info.options.get("database", None) or "??")
        if datasource_info.name:
            s = '%s (%s)' % (datasource_info.name, s)
        return s
    
    def dbconnect(self, data):
        opts = self._get_conn_opts_from_opts(data)
        if data.get("ask_for_password", False):
            pwd = self.password_prompt()
            if not pwd:
                raise DBConnectError(_(u"No password given."))
            opts["password"] = pwd
        try:
            import cx_Oracle
        except ImportError:
            raise DBConnectError(_(u"Python module cx_Oracle is not installed."))
        try:
            real_conn = cx_Oracle.connect(**opts)
        except cx_Oracle.OperationalError, err:
            raise DBConnectError(err.message)
        conn = OracleConnection(self.app, real_conn)
        conn.threadsafety = cx_Oracle.threadsafety
        return conn
    
    def test_connection(self, data):
        try:
            conn = self.dbconnect(data)
            conn.close()
        except DBConnectError, err:
            return err.message
        return None
    
class OracleCursor(DbAPI2Cursor):
    
    def __init__(self, *args, **kw):
        DbAPI2Cursor.__init__(self, *args, **kw)
        self._fetched = None
        
    def fetchall(self):
        if self._fetched == None:
            self._fetched = self._cur.fetchall()
        return self._fetched
    
    def execute(self, statement):
        DbAPI2Cursor.execute(self, statement)
        self.fetchall()
    
    def _get_rowcount(self):
        if self._fetched == None:
            return -1
        return len(self._fetched)
    rowcount = property(fget=_get_rowcount)
    
class OracleConnection(DbAPI2Connection):
    cursor_class = OracleCursor
    
    def get_server_info(self):
        return "Oracle %s" % self._conn.version
    
    
class OracleSchema(SchemaProvider):
    
    def q(self, connection, sql):
        cur = connection.cursor()._cur
        cur.execute(sql)
        return cur.fetchall()
    
    def fetch_children(self, connection, parent):
        if isinstance(parent, DatasourceInfo):
            return [TableCollection(),
                    ViewCollection(),
                    SequenceCollection(),
                    PackageCollection()]
        
        elif isinstance(parent, TableCollection):
            ret = []
            sql = "select t.table_name, c.comments \
            from user_tables t \
            left join user_tab_comments c on c.table_name = t.table_name \
            and c.table_type = 'TABLE'"
            for item in self.q(connection, sql):
                ret.append(Table(item[0], item[1]))
            return ret
        
        elif isinstance(parent, ViewCollection):
            ret = []
            sql = "select t.view_name, c.comments from user_views t \
            left join user_tab_comments c on c.table_name = t.view_name \
            and c.table_type = 'VIEW'"
            for item in self.q(connection, sql):
                ret.append(View(item[0]))
            return ret
        
        elif isinstance(parent, Table):
            return [ColumnCollection(table=parent),
                    ConstraintCollection(table=parent),
                    IndexCollection(table=parent)]
        
        elif isinstance(parent, View):
            return [ColumnCollection(table=parent)]
        
        elif isinstance(parent, ColumnCollection):
            table = parent.get_data("table")
            ret = []
            sql = "select col.column_name, com.comments \
            from user_tab_columns col \
            left join user_col_comments com on com.table_name = '%(table)s' and com.column_name = col.column_name \
            where col.table_name = '%(table)s'" % {"table" : table.name}
            [ret.append(Column(item[0], item[1])) for item in self.q(connection, sql)]
            return ret
        
        elif isinstance(parent, ConstraintCollection):
            table = parent.get_data("table")
            ret = []
            sql = "select constraint_name from user_constraints \
            where table_name = '%(table)s'" % {"table" : table.name}
            [ret.append(Constraint(item[0])) for item in self.q(connection, sql)]
            return ret
        
        elif isinstance(parent, IndexCollection):
            table = parent.get_data("table")
            ret = []
            sql = "select index_name from user_indexes \
            where table_name = '%(table)s'" % {"table" : table.name}
            [ret.append(Index(item[0])) for item in self.q(connection, sql)]
            return ret
        
        elif isinstance(parent, SequenceCollection):
            ret = []
            sql = "select sequence_name from user_sequences"
            [ret.append(Sequence(item[0])) for item in self.q(connection, sql)]
            return ret
        
        # FIXME: Cleanup packages and procedures
        elif isinstance(parent, PackageCollection):
            ret = []
            sql = "select distinct object_name from user_procedures \
            where object_name is not null"
            [ret.append(Package(item[0])) for item in self.q(connection, sql)]
            return ret
        
        elif isinstance(parent, Package):
            ret = []
            sql = "select procedure_name from user_procedures \
            where object_name = '%(package)s'" % {"package": parent.name}
            [ret.append(Function(item[0])) for item in self.q(connection, sql)]
            return ret


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: toolbar.py 16 2008-02-26 14:52:30Z albrecht.andi $

import gtk
import gobject

from gettext import gettext as _

from cf.backends import TRANSACTION_IDLE, TRANSACTION_COMMIT_ENABLED, TRANSACTION_ROLLBACK_ENABLED
from cf.ui import GladeWidget
from cf.ui.widgets import ConnectionButton

class CFToolbar(GladeWidget):
    
    def __init__(self, app, xml, cb_provider=None):
        GladeWidget.__init__(self, app, xml, "toolbar", cb_provider=cb_provider)
        item = self.xml.get_widget("tb_connection")
        self.tb_connection = ConnectionButton(self.app)
        item.add(self.tb_connection)
        self.__editor_signals = list()
        self.__conn_signals = list()
        self.set_editor(None)
        
    def on_connection_notify(self, connection, property):
        if property.name == "transaction-state":
            value = connection.get_property(property.name)
            gobject.idle_add(self.set_transaction_state, value)
        
    def on_editor_connection_changed(self, editor, conn):
        if editor != self._editor: return
        self.xml.get_widget("tb_execute").set_sensitive(bool(conn))
        if conn:
            self.set_transaction_state(conn.get_property("transaction-state"))
            self.__conn_signals.append(conn.connect("notify", self.on_connection_notify))
        else:
            self.set_transaction_state(None)
            
    def set_transaction_state(self, value):
        if value == None: 
            for item in ("tb_commit", "tb_rollback", "tb_begin"):
                self.xml.get_widget(item).set_sensitive(False)
                return
        self.xml.get_widget("tb_commit").set_sensitive((value & TRANSACTION_COMMIT_ENABLED) != 0)
        self.xml.get_widget("tb_rollback").set_sensitive((value & TRANSACTION_ROLLBACK_ENABLED) != 0)
        self.xml.get_widget("tb_begin").set_sensitive((value & TRANSACTION_IDLE) != 0)
        
    def set_editor(self, editor):
        while self.__editor_signals:
            self._editor.disconnect(self.__editor_signals.pop())
        self._editor = editor
        self.tb_connection.set_editor(editor)
        for item in ["tb_cut", "tb_copy", "tb_paste"]:
            self.xml.get_widget(item).set_sensitive(bool(editor))
        if self._editor:
            self.__editor_signals.append(self._editor.connect("connection-changed", self.on_editor_connection_changed))
            self.on_editor_connection_changed(self._editor, self._editor.connection)


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: browser.py 16 2008-02-26 14:52:30Z albrecht.andi $

"""Object browser"""

import gtk
import gobject

from cf.datasources import DatasourceInfo

from gettext import gettext as _

class DummyNode(object):
    pass

class Browser(gtk.ScrolledWindow):
    
    def __init__(self, app, instance):
        self.app = app
        self.instance = instance
        self._setup_widget()
        self._setup_connections()
        self.reset_tree()
        
    def _setup_widget(self):
        gtk.ScrolledWindow.__init__(self)
        self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.set_size_request(250, -1)
        # Model
        self.model = gtk.TreeStore(gobject.TYPE_PYOBJECT, # 0 Object
                                   str, # 1 label
                                   gtk.gdk.Pixbuf, # 2 icon
                                   gobject.TYPE_PYOBJECT, # 3 double-click callback
                                   gobject.TYPE_PYOBJECT, # 4 popup menu callback
                                   str, # 5 tooltip
                                   )
        self.model.set_sort_column_id(1, gtk.SORT_ASCENDING)
        # TreeView
        self.object_tree = gtk.TreeView(self.model)
        self.add(self.object_tree)
        self.object_tree.set_tooltip_column(5)
        col = gtk.TreeViewColumn()
        renderer = gtk.CellRendererPixbuf()
        col.pack_start(renderer, expand=False)
        col.add_attribute(renderer, 'pixbuf', 2)
        renderer = gtk.CellRendererText()
        col.pack_start(renderer, expand=True)
        col.add_attribute(renderer, 'text', 1)
        self.object_tree.append_column(col)
        self.object_tree.set_headers_visible(False)
        
    def _setup_connections(self):
        self.app.datasources.connect("datasource-added", self.on_datasource_added)
        self.app.datasources.connect("datasource-deleted", self.on_datasource_deleted)
        self.app.datasources.connect("datasource-modified", self.on_datasource_modified)
        self.object_tree.connect("button-press-event", self.on_button_press_event)
        self.object_tree.connect("row-expanded", self.on_row_expanded)
        sel = self.object_tree.get_selection()
        sel.connect("changed", self.on_object_tree_selection_changed)
        
    def _create_dsinfo_menu(self, model, iter, popup):
        def connect(item, datasource_info):
            datasource_info.dbconnect()
            self.on_object_tree_selection_changed(self.object_tree.get_selection())
        def disconnect(item, datasource_info):
            datasource_info.dbdisconnect()
        datasource_info = model.get_value(iter, 0)
        if datasource_info.get_connections():
            lbl = _(u"Disconnect")
            cb = disconnect
        else:
            lbl = _(u"Connect")
            cb = connect
        item = gtk.MenuItem(lbl)
        item.connect("activate", cb, datasource_info)
        popup.append(item)
        return popup
        
    def on_button_press_event(self, treeview, event):
        if event.button == 3:
            x = int(event.x)
            y = int(event.y)
            time = event.time
            pthinfo = treeview.get_path_at_pos(x, y)
            if pthinfo is not None:
                path, col, cellx, celly = pthinfo
                treeview.grab_focus()
                treeview.set_cursor( path, col, 0)
                popup = gtk.Menu()
                model = treeview.get_model()
                iter = model.get_iter(path)
                cb = model.get_value(iter, 4)
                if not cb:
                    return 1
                try:
                    popup = cb(model, iter, popup)
                except:
                    import traceback; traceback.print_exc()
                    return 1
                if popup.get_children():
                    popup.show_all()
                    popup.popup( None, None, None, event.button, time)
            return 1
        
    def on_datasource_added(self, manager, datasource_info):
        iter = self.model.append(None)
        self.set_datasource_info(iter, datasource_info)
    
    def on_datasource_deleted(self, manager, datasource_info):
        iter = self.get_iter_for_datasource(datasource_info)
        if not iter:
            return
        self.model.remove(iter)
    
    def on_datasource_modified(self, manager, datasource_info):
        iter = self.get_iter_for_datasource(datasource_info)
        if not iter:
            return
        self.set_datasource_info(iter, datasource_info)
        
    def on_object_tree_selection_changed(self, selection):
        model, iter = selection.get_selected()
        if not iter: return
        obj = model.get_value(iter, 0)
        if isinstance(obj, DatasourceInfo):
            if obj.internal_connection:
                server_info = obj.internal_connection.get_server_info()
                if server_info:
                    self.instance.statusbar.set_message(server_info)
                else:
                    self.instance.statusbar.set_message("")
            else:
                self.instance.statusbar.set_message("")
        else:
            comm = model.get_value(iter, 5) or ""
            self.instance.statusbar.set_message(comm)
            
        
    def on_row_expanded(self, treeview, iter, path):
        model = treeview.get_model()
        obj = model.get_value(iter, 0)
        citer = model.iter_children(iter)
        if citer:
            cobj = model.get_value(citer, 0)
            if isinstance(cobj, DummyNode):
                datasource_info = self.find_datasource_info(model, citer)
                model.remove(citer)
                if datasource_info.backend.schema:
                    for child in datasource_info.backend.schema.fetch_children(datasource_info.internal_connection, obj) or []:
                        citer = model.append(iter)
                        if child.icon:
                            icon = gtk.icon_theme_get_default().load_icon(child.icon, gtk.ICON_SIZE_MENU, gtk.ICON_LOOKUP_FORCE_SVG)
                        else:
                            icon = None
                        model.set(citer, 
                                  0, child,
                                  1, child.name,
                                  2, icon,
                                  5, child.description)
                        if child.has_children:
                            cciter = model.append(citer)
                            model.set(cciter, 0, DummyNode())
                    citer = model.iter_children(iter)
                    if citer:
                        treeview.expand_row(model.get_path(iter), False)
                    
    def find_datasource_info(self, model, iter):
        while iter:
            obj = model.get_value(iter, 0)
            if isinstance(obj, DatasourceInfo):
                return obj
            iter = model.iter_parent(iter)
    
    def get_iter_for_datasource(self, datasource_info):
        iter = self.model.get_iter_first()
        while iter:
            if self.model.get_value(iter, 0) == datasource_info:
                return iter
            iter = self.model.iter_next(iter)
        return None
    
    def reset_tree(self):
        self.model.clear()
        for datasource_info in self.app.datasources.get_all():
            iter = self.model.append(None)
            self.set_datasource_info(iter, datasource_info)
    
    def set_datasource_info(self, iter, datasource_info):
        if datasource_info.get_connections():
            ico = gtk.icon_theme_get_default().load_icon("stock_connect", gtk.ICON_SIZE_MENU, gtk.ICON_LOOKUP_FORCE_SVG)
        else:
            ico = gtk.icon_theme_get_default().load_icon("stock_disconnect", gtk.ICON_SIZE_MENU, gtk.ICON_LOOKUP_FORCE_SVG)
        self.model.set(iter,
                  0, datasource_info,
                  1, datasource_info.get_label(),
                  2, ico,
                  3, None,
                  4, self._create_dsinfo_menu,
                  5, datasource_info.description)
        citer = self.model.iter_children(iter)
        if datasource_info.get_connections() \
        and not citer:
            citer = self.model.append(iter)
            self.model.set(citer, 0, DummyNode())
        elif not datasource_info.get_connections() \
        and citer:
            self.model.remove(citer)


name = "CrunchyFrog"
appname = name.lower()
version = "0.2"
copyright = "Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>"
description = "Simple GNOME database tool"
author = "Andi Albrecht"
author_email = "albrecht.andi@gmail.com"
url = "http://crunchyfrog.googlecode.com"
long_description = """"CrunchyFrog is a SQL editor and database schema browser for the GNOME desktop.

Supported databases
===================

 * PostgreSQL (requires psycopg2)
 * MySQL (requires MySQLdb)
 * SQLite (requires python-sqlite3)
 * Oracle (requires cx_Oracle)

It is written in Python (PyGTK) and provides a setuptools based plugin system, which is completely undocumented for now (read the source ;-)

Check out the Google pages for the most recent version."""
license = """This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>."""


# -*- coding: utf-8 -*-

# pocdock - an experimental docking implementation
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""An experimental docking implementation.
    
# Todo
 * API cleanup
 
"""

import gtk
import gobject
from lxml import etree
from cStringIO import StringIO
import cairo

DOCK_ITEM_STATUS_DOCKED = 'docked'
DOCK_ITEM_STATUS_HIDDEN = 'hidden'
DOCK_ITEM_STATUS_FLOATING = 'floating'

DOCK_ITEM_BEH_NORMAL = 0
DOCK_ITEM_BEH_NEVER_FLOATING = 1 << 0
DOCK_ITEM_BEH_NEVER_VERTICAL = 1 << 1
DOCK_ITEM_BEH_NEVER_HORIZONTAL = 1 << 2
DOCK_ITEM_BEH_CANT_CLOSE = 1 << 3
DOCK_ITEM_BEH_CANT_AUTOHIDE = 1 << 4
DOCK_ITEM_BEH_LOCKED = 1 << 5


class Dock(gtk.VBox):
    """The docking widget."""
    
    def __init__(self):
        self.__gobject_init__()
        self.items = list()
        self.bar_top = DockBar(self, gtk.POS_TOP)
        self.bar_bottom = DockBar(self, gtk.POS_BOTTOM)
        self.bar_left = DockBar(self, gtk.POS_LEFT)
        self.bar_right = DockBar(self, gtk.POS_RIGHT)
        self.pack_start(self.bar_top, False, False, 0)
        hbox = gtk.HBox()
        hbox.pack_start(self.bar_left, False, False, 0)
        self.main = DockGroup(self, "__main__")
        hbox.pack_start(self.main, True, True, 0)
        hbox.pack_start(self.bar_right, False, False, 0)
        self.pack_start(hbox, True, True, 0)
        self.pack_start(self.bar_bottom, False, False, 0)
        
    def add_item(self, item):
        self.main.add_item(item)
        
    def find_group_by_position(self, dock_item, x, y):
        gx, gy = self.main.window.get_origin()
        if x >= gx and x <= gx+self.main.allocation.width \
        and y >= gy and y <= gy+self.main.allocation.width:
            return self.main.find_group_by_position(dock_item, x, y)
        else:
            return None, None
            
    def flatten_groups(self):
        return self.main.flatten_groups()
    
    def update_views(self):
        for group in self.flatten_groups():
            group.update_views()
                
    def save_layout(self, fp):
        dom = etree.Element("dock-layout")
        self.main.save_layout(dom)
        for item in self.items:
            if item.status == DOCK_ITEM_STATUS_FLOATING:
                win = item.get_parent()
                width = win.allocation.width
                height = win.allocation.height
                x, y = win.get_position()
                fl = etree.Element("floating")
                fl.set("x", str(x))
                fl.set("y", str(y))
                fl.set("width", str(width))
                fl.set("height", str(height))
                x = etree.Element("item")
                x.set("id", item.id)
                fl.append(x)
                dom.append(fl)
        fp.write(etree.tounicode(dom, pretty_print=True))
        
    def dump(self):
        fp = StringIO()
        self.save_layout(fp)
        print fp.getvalue()
        
        
class DockBar(gtk.HBox):
    """Container for *hidden* DockItems."""
    
    def __init__(self, dock, pos):
        gtk.HBox.__init__(self)
        self.dock = dock
        self.bar = gtk.Toolbar()
        self.pos = pos
        if pos in [gtk.POS_LEFT, gtk.POS_RIGHT]:
            self.bar.set_orientation(gtk.ORIENTATION_VERTICAL)
        else:
            self.bar.set_orientation(gtk.ORIENTATION_HORIZONTAL)
        self.pack_start(self.bar, True, True)
            
    def add_item(self, item):
        btn = gtk.ToolButton()
        btn.set_data("item", item)
        lbl = gtk.Label(item.label)
        btn.set_label_widget(lbl)
        if self.pos in [gtk.POS_LEFT, gtk.POS_RIGHT]:
            lbl.set_angle(270)
        self.bar.insert(btn, -1)
        self.show_all()
        btn.connect("clicked", self.on_button_clicked)
        
    def on_button_clicked(self, btn):
        btn.get_data("item").show()
        btn.get_parent().remove(btn)
        btn.destroy()
        if not self.bar.get_n_items():
            self.hide()
            
class DockNotebook(gtk.Notebook):
    """The main view within a DockGroup"""
    
    def __init__(self):
        self.__gobject_init__()
        self.set_tab_pos(gtk.POS_BOTTOM)
        self.set_scrollable(True)
        self.popup_enable()
        self.connect("page-removed", self.on_page_count_changed)
        self.connect("page-added", self.on_page_count_changed)
        
    def on_page_count_changed(self, *args):
        self.set_show_tabs(self.get_n_pages() > 1)
        self.set_show_border(self.get_n_pages() > 1)


class DockGroup(gtk.EventBox):
    """A DockGroup holds a centered main view and up to four
    DockGroups (top, bottom, right, left)"""
    
    def __init__(self, dock, name):
        self.__gobject_init__()
        self.dock = dock
        self.grpname = name
        self.objects = list()
        self.vpaned1 = gtk.VPaned()
        self.add(self.vpaned1)
        self.vpaned2 = gtk.VPaned()
        self.vpaned1.add2(self.vpaned2)
        self.hpaned1 = gtk.HPaned()
        self.vpaned2.add1(self.hpaned1)
        self.hpaned2 = gtk.HPaned()
        self.hpaned1.add2(self.hpaned2)
        self.notebook = DockNotebook()
        self.hpaned2.add1(self.notebook)
        self.group_top = None
        self.group_bottom = None
        self.group_left = None
        self.group_right = None
        self.connect("remove", self.on_remove)
        
    def _get_subgroups(self):
        r = list()
        for item in ["top", "bottom", "left", "right"]:
            group = getattr(self, "group_%s" % item, None)
            if group:
                r.append(group)
        return r
    subgroups = property(fget=_get_subgroups)
        
    def on_remove(self, *args):
        if not self.group_bottom and not self.group_top \
        and not self.group_left and not self.group_right \
        and not self.notebook.get_n_pages() \
        and self.get_parent():
            self.get_parent().remove(self)
            self.destroy()
            return
            
    def on_group_destroyed(self, widget, group):
        gobject.idle_add(setattr, self, group, None)
        
    def flatten_groups(self):
        groups=[]
        if self.group_bottom:
            groups += self.group_bottom.flatten_groups()
        if self.group_top:
            groups += self.group_top.flatten_groups()
        if self.group_left:
            groups += self.group_left.flatten_groups()
        if self.group_right:
            groups += self.group_right.flatten_groups()
        groups.append(self)
        return groups
    
    def find_group_by_position(self, dock_item, x, y):
        for group in self.subgroups:
            gx, gy = group.window.get_origin()
            if x >= gx and x <= gx+group.allocation.width \
            and y >= gy and y <= gy+group.allocation.height:
                return group.find_group_by_position(dock_item, x, y)
        gx, gy = self.window.get_origin()
        if y <= gy+self.allocation.height/3.0 \
        and dock_item.behavior & DOCK_ITEM_BEH_NEVER_VERTICAL == 0:
            pos = gtk.POS_TOP
        elif y >= gy+self.allocation.height-self.allocation.height/3.0 \
        and dock_item.behavior & DOCK_ITEM_BEH_NEVER_VERTICAL == 0:
            pos = gtk.POS_BOTTOM
        elif x <= gx+self.allocation.width/3.0 \
        and dock_item.behavior & DOCK_ITEM_BEH_NEVER_HORIZONTAL == 0:
            pos = gtk.POS_LEFT
        elif x >= gx+self.allocation.width-self.allocation.width/3.0 \
        and dock_item.behavior & DOCK_ITEM_BEH_NEVER_HORIZONTAL == 0:
            pos = gtk.POS_RIGHT
        else:
            pos = None
        return self, pos
        
    def add_item(self, item, pos=-1):
        if pos == -1:
            pos = item.initial_position
        else:
            pos = pos
        if pos == None:
            item.status = DOCK_ITEM_STATUS_DOCKED
            self.notebook.append_page(item, gtk.Label(item.label))
            self.notebook.set_current_page(self.notebook.get_n_pages())
            self.notebook.set_tab_reorderable(item, True)
        elif pos == gtk.POS_TOP:
            if self.group_top:
                self.group_top.add_item(item, pos=None)
            else:
                self.group_top = DockGroup(self.dock, "top")
                self.group_top.add_item(item, pos=None)
                self.vpaned1.add1(self.group_top)
                self.group_top.connect("destroy", self.on_group_destroyed, "group_top")
        elif pos == gtk.POS_BOTTOM:
            if self.group_bottom:
                self.group_bottom.add_item(item, pos=None)
            else:
                self.group_bottom = DockGroup(self.dock, "bottom")
                self.group_bottom.add_item(item, pos=None)
                self.vpaned2.add2(self.group_bottom)
                self.group_bottom.connect("destroy", self.on_group_destroyed, "group_bottom")
        elif pos == gtk.POS_LEFT:
            if self.group_left:
                self.group_left.add_item(item, pos=None)
            else:
                self.group_left = DockGroup(self.dock, "left")
                self.group_left.add_item(item, pos=None)
                self.hpaned1.add1(self.group_left)
                self.group_left.connect("destroy", self.on_group_destroyed, "group_left")
        elif pos == gtk.POS_RIGHT:
            if self.group_right:
                self.group_right.add_item(item, pos=None)
            else:
                self.group_right = DockGroup(self.dock, "right")
                self.group_right.add_item(item, pos=None)
                self.hpaned2.add2(self.group_right)
                self.group_right.connect("destroy", self.on_group_destroyed, "group_right")
        self.show_all()
        self.dock.update_views()
        
    def relocate_item(self, item, pos):
        parent = item.get_parent()
        if isinstance(parent, gtk.Notebook):
            for n in range(parent.get_n_pages()):
                if parent.get_nth_page(n) == item:
                    parent.remove_page(n)
                    break
        elif item.get_parent():
            item.get_parent().remove(item)
            item.destroy()
        self.add_item(item, pos)
        
    def has_items(self):
        if self.notebook.get_n_pages() != 0:
            return True
        if self.group_top \
        and self.group_top.has_items():
            return True
        if self.group_bottom \
        and self.group_bottom.has_items():
            return True
        if self.group_left \
        and self.group_left.has_items():
            return True
        if self.group_right \
        and self.group_right.has_items():
            return True
        return False

    def update_views(self):
        if self.group_top and not self.group_top.has_items():
            self.group_top.destroy()
        if self.group_bottom and not self.group_bottom.has_items():
            self.group_bottom.destroy()
        if self.group_left and not self.group_left.has_items():
            self.group_left.destroy()
        if self.group_right and not self.group_right.has_items():
            self.group_right.destroy()
        group_count = 0
        last_group = None
        if self.group_top:
            group_count += 1
            last_group = self.group_top
        if self.group_bottom:
            group_count += 1
            last_group = self.group_bottom
        if self.group_right:
            group_count += 1
            last_group = self.group_right
        if self.group_left:
            group_count += 1
            last_group = self.group_left
        if group_count == 1 and not self.notebook.get_n_pages():
            items = list()
            while last_group.notebook.get_n_pages():
                item = last_group.notebook.get_nth_page(0)
                last_group.notebook.remove_page(0)
                items.append(item)
            for item in items:
                self.add_item(item, pos=None)
            
    
    def show_dock_target(self, win, pos):
        x = y = width = height = None
        dx, dy = self.window.get_origin()
        if pos == gtk.POS_TOP:
            x = dx
            y = dy
            width = self.allocation.width
            height = self.allocation.height / 3
        elif pos == gtk.POS_BOTTOM:
            x = dx
            y = dy + self.allocation.height - self.allocation.height/3
            width = self.allocation.width
            height = self.allocation.height / 3
        elif pos == gtk.POS_LEFT:
            x = dx
            y = dy
            width = self.allocation.width / 3
            height = self.allocation.height
        elif pos == gtk.POS_RIGHT:
            x = dx + self.allocation.width - self.allocation.width / 3
            y = dy
            width = self.allocation.width / 3
            height = self.allocation.height
        elif pos == None:
            x = dx + self.allocation.width / 3
            y = dy + self.allocation.height / 3
            width = self.allocation.width / 3
            height = self.allocation.height / 3
        if width and height:
            win.relocate(x, y, width, height, True)
            win.show()
        else:
            win.hide()
            
    def save_layout(self, node, pos=None):
        grp = etree.Element("group")
        if pos:
            grp.set("pos", pos)
            grp.set("position", str(self.get_parent().get_position()))
        grp.set("name", self.grpname)
        node.append(grp)
        for n in range(self.notebook.get_n_pages()):
            item = self.notebook.get_nth_page(n)
            xitem = etree.Element("item")
            xitem.set("id", str(item.id))
            grp.append(xitem)
        if self.group_top:
            self.group_top.save_layout(grp, "top")
        if self.group_bottom:
            self.group_bottom.save_layout(grp, "bottom")
        if self.group_left:
            self.group_left.save_layout(grp, "left")
        if self.group_right:
            self.group_right.save_layout(grp, "right")
        
class DockItem(gtk.VBox):
    """A DockItem"""
    
    def __init__(self, dock, id, widget, label, icon=None, pos=None,
                 behavior=DOCK_ITEM_BEH_NORMAL):
        self.__gobject_init__()
        self.id = id
        self.dock = dock
        self.widget = widget
        self.label = label
        self.icon = icon
        self.behavior = behavior
        self.shadow_window = None
        self.status = None
        self.initial_position = pos
        self._setup_widget()
        
    def _setup_widget(self):
        self.header = gtk.EventBox()
        style = self.get_style()
        self.header.modify_bg(gtk.STATE_NORMAL, style.bg[gtk.STATE_ACTIVE])
        hbox = gtk.HBox()
        hbox.set_spacing(3)
        if self.icon:
            i = gtk.image_new_from_stock(self.icon, gtk.ICON_SIZE_MENU)
            hbox.pack_start(i, False, False)
        a = gtk.Alignment(0, 0.5)
        a.add(gtk.Label(self.label))
        hbox.pack_start(a, True, True, 0)
        if self.behavior & DOCK_ITEM_BEH_LOCKED == 0:
            if self.behavior & DOCK_ITEM_BEH_CANT_AUTOHIDE == 0:
                btn = gtk.Button()
                btn.set_relief(gtk.RELIEF_NONE)
                btn.set_property("can-focus", False)
                btn.set_size_request(17, 17)
                btn.set_image(gtk.image_new_from_stock("gtk-leave-fullscreen", gtk.ICON_SIZE_MENU))
                btn.connect("clicked", self.on_send_to_bar)
                hbox.pack_start(btn, False, False)
            if self.behavior & DOCK_ITEM_BEH_CANT_CLOSE == 0:
                btn = gtk.Button()
                btn.set_relief(gtk.RELIEF_NONE)
                btn.set_property("can-focus", False)
                btn.set_size_request(17, 17)
                btn.set_image(gtk.image_new_from_stock("gtk-close", gtk.ICON_SIZE_MENU))
                btn.connect("clicked", self.on_close)
                hbox.pack_start(btn, False, False)
        self.header.add(hbox)
        if self.behavior & DOCK_ITEM_BEH_LOCKED == 0:
            self.header.connect("enter-notify-event", self.on_header_enter)
            self.header.connect("leave-notify-event", self.on_header_leave)
            self.header.connect("button-press-event", self.on_header_button_press)
            self.header.connect("motion-notify-event", self.on_header_motion)
            self.header.connect("button-release-event", self.on_header_button_release)
        self.header.connect("realize", self.on_header_realize)
        self.pack_start(self.header, False, False)
        if self.widget.get_parent():
            self.widget.reparent(self)
            self.set_child_packing(self.widget, True, True, 0, gtk.PACK_START)
        else:
            self.pack_start(self.widget, True, True)
        self.dock.items.append(self)
        self.connect("destroy", self.on_destroy)
            
    def on_close(self, *args):
        self.destroy()
        
    def on_destroy(self, *args):
        self.dock.items.remove(self)
        
    def on_send_to_bar(self, *args):
        self.send_to_bar()
        
    def on_header_enter(self, *args):
        style = self.get_style()
        self.header.modify_bg(gtk.STATE_NORMAL, style.bg[gtk.STATE_PRELIGHT])
        
    def on_header_leave(self, *args):
        style = self.get_style()
        self.header.modify_bg(gtk.STATE_NORMAL, style.bg[gtk.STATE_ACTIVE])
        
    def on_header_button_press(self, widget, event):
        if event.button == 1:
            self.shadow_window = PlaceholderWindow(self.dock, self)
            self.header.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.FLEUR))
            
    def on_header_button_release(self, widget, event):
        if event.button == 1:
            x, y = event.get_root_coords()
            group, pos = self.dock.find_group_by_position(self, x, y)
            if group:
                group.relocate_item(self, pos)
            else:
                self.set_floating(x, y)
            self.shadow_window.destroy()
            self.shadow_window = None
            gobject.idle_add(self.dock.update_views)
            
    def on_header_motion(self, widget, event):
        x, y = event.get_root_coords()
        group, pos = self.dock.find_group_by_position(self, x, y)
        if group:
            gobject.idle_add(group.show_dock_target, self.shadow_window, pos)
        elif self.behavior & DOCK_ITEM_BEH_NEVER_FLOATING == 0:
            gobject.idle_add(self.shadow_window.relocate, x, y, self.allocation.width, self.allocation.height, True)
            
    def on_header_realize(self, header, *args):
        header.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.HAND2))
        
    def send_to_bar(self):
        self.hide()
        parent = self.get_parent()
        topgroup = None
        while parent:
            if isinstance(parent, DockGroup) \
            and parent.grpname != "__main__":
                topgroup = parent
            parent = parent.get_parent()
        if not topgroup or topgroup.grpname == "top":
            bar = self.dock.bar_top
        elif topgroup.grpname == "left":
            bar = self.dock.bar_left
        elif topgroup.grpname == "right":
            bar = self.dock.bar_right
        else:
            bar = self.dock.bar_bottom
        bar.add_item(self)
        
    def set_floating(self, x, y):
        width = self.allocation.width
        height = self.allocation.height
        w = FloatingWindow(self)
        w.resize(width, height)
        w.move(int(x), int(y))
        w.show_all()
        self.status = DOCK_ITEM_STATUS_FLOATING

# Parts of the PlaceholderWindow are ported from MonoDevelop's
# docking implementation.       
class PlaceholderWindow(gtk.Window):
    """A *dummy* window which is displayed when dragging around DockItems"""
    redgc = None
    anim = 0
    rx = 0
    ry = 0
    rw = 0
    rh = 0
    allow_docking = False
    
    with_preview = False
    
    def __init__(self, dock, dockitem):
        gtk.Window.__init__(self, gtk.WINDOW_POPUP)
        self.use_alpha = False
        self.set_skip_pager_hint(True)
        self.set_decorated(False)
        self.set_transient_for(dock.get_toplevel())
        self.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_UTILITY)
        self._set_colormap()
        if self.with_preview:
            pb = gtk.gdk.pixbuf_get_from_drawable(None,
                        dockitem.window, dockitem.window.get_colormap(),
                        0, 0, 0, 0, dockitem.allocation.width, dockitem.allocation.height)
            f1 = 150.0/dockitem.allocation.width
            f2 = 250.0/dockitem.allocation.height
            if f1 < f2:
                w = dockitem.allocation.width*f1
                h = dockitem.allocation.height*f1
            else:
                w = dockitem.allocation.width*f2
                h = dockitem.allocation.height*f2
            pb = pb.scale_simple(int(w), int(h), gtk.gdk.INTERP_NEAREST)
            img = gtk.Image()
            img.set_from_pixbuf(pb)
            a = gtk.Alignment(0, 0)
            a.add(img)
            self.add(a)
            self.show_all()
        self.set_app_paintable(True)
        self.realize()
        if self.is_composited():
            self.connect("expose-event", self.expose)
        else:
            self.connect("expose-event", self.expose_simple)
        
    def _set_colormap(self):
        screen = self.get_screen()
        colormap = screen.get_rgba_colormap()
        self.use_alpha = True
        if not colormap:
            colormap = screen.get_rgb_colormap()
            self.use_alpha = False
        self.set_colormap(colormap)

    def relocate(self, x, y, w, h, animate):
        x = int(x)
        y = int(y)
        if x != self.rx or y != self.ry or w != self.rw or h != self.rh:
            self.move(x, y)
            self.resize(w, h)
            
            self.rx = x
            self.ry = y
            self.rw = w
            self.rh = h
            
            if self.anim != 0:
                gobject.source_remove(self.anim)
                self.anim = 0
            if animate and w < 150 and h < 150:
                sa = 7
                self.move(self.rx-sa, self.ry-sa)
                self.resize(self.rw+sa*2, self.rh+sa*2)
                self.anim = gobject.timeout_add(10, self.run_animation)

    def run_animation(self):
        cw, ch = self.get_size()
        cx, cy = self.get_position()
        if cx != self.rx:
            cx += 1
            cy += 1
            ch -= 2
            cw -= 2
            self.move(cx, cy)
            self.resize(cw, ch)
            return True
        self.anim = 0
        return False 
    
    def draw(self, widget, cr):
        rect = widget.get_allocation()
        draw_rb(cr, rect.x, rect.y, rect.width, rect.height, 5)
        cr.set_operator(cairo.OPERATOR_OVER)
        cr.set_source_rgba(0, 0, 0, .2)
        cr.fill_preserve()
            
    def expose(self, widget, event):
        cr = widget.window.cairo_create()
        cr.rectangle(event.area.x, event.area.y,
                     event.area.width, event.area.height)
        cr.clip()
        cr.set_source_rgba(0, 0, 0, .2)
        cr.set_operator( cairo.OPERATOR_SOURCE )
        cr.paint()
        self.draw(widget, cr)
        return False
    
    def expose_simple(self, widget, event):
        cr = widget.window.cairo_create()
        cr.rectangle(event.area.x, event.area.y,
                     event.area.width, event.area.height)
        cr.clip()
        cr.set_operator( cairo.OPERATOR_SOURCE )
        cr.set_source_rgb(0, 0, 0)
        width = self.allocation.width
        height = self.allocation.height
        cr.move_to(0, 0)
        cr.rel_line_to(0, height)
        cr.rel_line_to(width, 0)
        cr.rel_line_to(0, -height)
        cr.rel_line_to(-width, 0)
        cr.stroke()

# draw_cb is ported from GNOME Launch Box 
def draw_rb(cr, x, y, width, height, radius):
    import math
    width -= 2 * radius
    height -= 2 * radius

    cr.move_to(x + radius, y)

    cr.rel_line_to(width, 0.0)

    cx, cy = cr.get_current_point()
    cr.arc (cx, cy + radius, radius, 3.0 * math.pi/2, 0)

    cr.rel_line_to (0.0, height)

    cx, cy = cr.get_current_point ()
    cr.arc (cx - radius, cy, radius, 0, math.pi/2)

    cr.rel_line_to (- width, 0.0)

    cx, cy = cr.get_current_point()
    cr.arc (cx, cy - radius, radius, math.pi/2, math.pi)

    cr.rel_line_to (0.0, -height)

    cx, cy = cr.get_current_point ()
    cr.arc (cx + radius, cy, radius, math.pi, 3.0 * math.pi/2)

    cr.close_path()
    
class FloatingWindow(gtk.Window):
    """A window for floating DockItems"""
    
    def __init__(self, item):
        gtk.Window.__init__(self)
        self.item = item
        self.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_UTILITY)
        self.item.reparent(self)
        self.connect("remove", self.on_remove)
        
    def on_remove(self, *args):
        if not self.get_child():
            self.destroy()


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: __init__.py 4 2008-02-26 05:04:22Z albrecht.andi $

import bonobo
import gnome
import gnomevfs
import gobject
import gtk
import gtk.glade
import gettext

from optparse import OptionParser
from os.path import abspath, dirname, join, isfile, isdir, expanduser
from os import makedirs
import sys
import logging

from cf import release

if isfile(abspath(join(dirname(__file__), "../setup.py"))):
    DATA_DIR = abspath(join(dirname(__file__), "../data"))
    LOCALE_DIR = abspath(join(dirname(__file__), "../po"))
else:
    root = abspath(join(dirname(sys.argv[0]), "../"))
    DATA_DIR = join(root, "share", release.appname)
    LOCALE_DIR = join(root, "share", "locale")
PLUGIN_DIR = join(DATA_DIR, "plugins")
USER_CONFIG_DIR = abspath(expanduser("~/.config/crunchyfrog"))
USER_CONF = join(USER_CONFIG_DIR, "config")
if not isdir(USER_CONFIG_DIR):
    makedirs(USER_CONFIG_DIR)
USER_DIR = abspath(expanduser("~/.crunchyfrog"))
if not isdir(USER_DIR):
    makedirs(USER_DIR)
USER_PLUGIN_DIR = join(USER_DIR, "plugins/")
if not isdir(USER_PLUGIN_DIR):
    makedirs(USER_PLUGIN_DIR)
USER_PLUGIN_URI = gnomevfs.get_uri_from_local_path(USER_PLUGIN_DIR)
    
gettext.bindtextdomain("crunchyfrog", LOCALE_DIR)
gettext.textdomain("crunchyfrog")
gtk.glade.bindtextdomain("crunchyfrog", LOCALE_DIR)
gtk.glade.textdomain("crunchyfrog")

if sys.version_info[:2] >= (2, 5):
    LOG_FORMAT_APP = '[%(levelname)s] %(message)s\n\t%(pathname)s:%(lineno)s in %(funcName)s\n\tPID: %(process)s, Thread: %(threadName)s [%(thread)d]\n'
else:
    LOG_FORMAT_APP = '[%(levelname)s] %(message)s\n\t%(pathname)s:%(lineno)s\n\tPID: %(process)s, Thread: %(threadName)s [%(thread)d]\n'


from cf.app import CFApplication

def new_instance_cb(xapp, argc, argv):
    from cf.instance import CFInstance
    app = CFInstance(xapp)
    app._init_ui(argv)
    l = xapp.get_data("instances")
    l.append(app)
    xapp.set_data("instances", l)
    app.widget.show_all()
    app.widget.connect("destroy", ui_destroy_cb, app, xapp)
    return argc

def ui_destroy_cb(widget, app, xapp):
    l = xapp.get_data("instances")
    if app in l:
        l.remove(app)
    xapp.set_data("instances", l)
    if not l:
        xapp.shutdown()
        bonobo.main_quit()

def _parse_commandline():
    """Parses command line arguments and handles all arguments
    which exit immediately (e.g. --version)"""
    usage = "usage: %prog [options] FILE1, FILE2, ..."
    parser = OptionParser(usage)
    parser.add_option("-d", "--debug",
                      action="store_true", dest="debug",
                      default=False,
                      help="run in debug mode")
    parser.add_option("--version",
                      action="store_true", dest="show_version",
                      default=False,
                      help="show program's version number and exit")
    options, args = parser.parse_args()
    if options.show_version:
        print "%s - %s" % (release.name, release.description)
        print "Version %s" % release.version
        print
        sys.exit()
    return options, args

def main():
    options, args = _parse_commandline()
    if options.debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.WARNING
    logging.basicConfig(format=LOG_FORMAT_APP,
                        level=log_level)
    bonobo.activate()
    app = CFApplication()
    client = app.register_unique(app.create_serverinfo(("LANG",)))
    if not client:
        if isfile(abspath(join(dirname(__file__), "../setup.py"))):
            props = {'app-datadir': abspath(join(dirname(__file__), '../data'))}
        else:
            props = dict()
        gnome.init(release.name.lower(), release.version,
                   properties=props)
        app.connect("new-instance", new_instance_cb)
        app.set_data("instances", list())
        app.new_instance(args)
        gobject.threads_init()
        gtk.gdk.threads_init()
        bonobo.main()
        app.unref()
    else:
        client.new_instance(args)
        
if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: editor.py 20 2008-02-26 20:18:01Z albrecht.andi $

import gtk
import gobject
import gconf
import pango

import re
import thread
import time

from gettext import gettext as _

from cf.backends import Query
from cf.ui import GladeWidget
from cf.ui.toolbar import CFToolbar

try:
    import gtksourceview2
    USE_GTKSOURCEVIEW2 = True
except ImportError:
    import gtksourceview
    USE_GTKSOURCEVIEW2 = False

class Editor(GladeWidget):
    
    __gsignals__ = {
        "connection-changed" : (gobject.SIGNAL_RUN_LAST,
                                gobject.TYPE_NONE,
                                (gobject.TYPE_PYOBJECT,))
    }
    
    def __init__(self, app, instance):
        self.app = app
        self.instance = instance
        self.connection = None
        self._query_timer = None
        GladeWidget.__init__(self, app, "crunchyfrog", "box_editor")
        self.set_data("win", None)
        self.show_all()
        
    def _setup_widget(self):
        self._setup_textview()
        self._setup_resultsgrid()
        self.lbl_status = self.xml.get_widget("editor_label_status")
        
    def _setup_textview(self):
        if USE_GTKSOURCEVIEW2:
            self.textview = gtksourceview2.View()
            buffer = gtksourceview2.Buffer()
            self.textview.set_buffer(buffer)
            lm = gtksourceview2.language_manager_get_default()
            buffer.set_language(lm.get_language("sql"))
        else:
            self.textview = gtksourceview.SourceView()
            buffer = gtksourceview.SourceBuffer()
            self.textview.set_buffer(buffer)
            buffer.set_highlight(True)
            lm = gtksourceview.SourceLanguagesManager()
            for lang in lm.get_available_languages():
                if lang.get_id() == "SQL":
                    buffer.set_language(lang)
                    break
        self.update_textview_options()
        sw = self.xml.get_widget("sw_editor_textview")
        sw.add(self.textview)
        
    def _setup_resultsgrid(self):
        self.results = ResultsView(self.app, self.instance, self.xml)
        
    def _setup_connections(self):
        self.textview.connect("populate-popup", self.on_populate_popup)
        self.app.config.connect("changed", self.on_config_changed)
        
    def on_close(self, *args):
        if self.get_data("win"):
            self.get_data("win").close()
        else:
            self.destroy()
            
    def on_config_changed(self, config, option, value):
        if option.startswith("editor."):
            gobject.idle_add(self.update_textview_options)
        
    def on_populate_popup(self, textview, popup):
        sep = gtk.SeparatorMenuItem()
        sep.show()
        popup.append(sep)
        if self.get_data("win"):
            lbl = _(u"Show in main window")
            cb = self.on_show_in_main_window
        else:
            lbl = _(u"Show in separate window")
            cb = self.on_show_in_separate_window
        item = gtk.MenuItem(lbl)
        item.connect("activate", cb)
        item.show()
        popup.append(item)
        item = gtk.ImageMenuItem("gtk-close")
        item.show()
        item.connect("activate", self.on_close)
        popup.append(item)
        
    def on_query_started(self, query):
        start = time.time()
        self._query_timer = gobject.timeout_add(10, self.update_exectime, start, query)
        
    def on_query_finished(self, query, tag_notice):
        self.results.set_query(query)
        self.connection.disconnect(tag_notice)
        
    def on_show_in_main_window(self, *args):
        gobject.idle_add(self.show_in_main_window)
        
    def on_show_in_separate_window(self, *args):
        gobject.idle_add(self.show_in_separate_window)
        
    def commit(self):
        if not self.connection: return
        cur = self.connection.cursor()
        cur.execute("commit")
        cur.close()
        
    def rollback(self):
        if not self.connection: return
        cur = self.connection.cursor()
        cur.execute("rollback")
        cur.close()
        
    def begin_transaction(self):
        if not self.connection: return
        cur = self.connection.cursor()
        cur.execute("begin")
        cur.close()
        
    def execute_query(self):
        self.lbl_status.set_text(_(u"Executing query..."))
        def exec_threaded(statement):
            cur = self.connection.cursor() 
            query = Query(statement, cur)
            gtk.gdk.threads_enter()
            query.connect("started", self.on_query_started)
            query.connect("finished", self.on_query_finished, tag_notice)
            gtk.gdk.threads_leave()
            query.execute(True)
        buffer = self.textview.get_buffer()
        bounds = buffer.get_selection_bounds()
        self.results.reset()
        def foo(connection, msg):
            self.results.add_message(msg)
        tag_notice = self.connection.connect("notice", foo)
        if not bounds:
            bounds = buffer.get_bounds()
        if self.connection.threadsafety >= 2:
            statement = buffer.get_text(*bounds)
            thread.start_new_thread(exec_threaded, (statement,))
        else:
            cur = self.connection.cursor()
            query = Query(buffer.get_text(*bounds), cur)
            query.connect("finished", self.on_query_finished, tag_notice)
            query.execute()
        
    def set_connection(self, conn):
        self.connection = conn
        self.emit("connection-changed", conn)
        
    def show_in_separate_window(self):
        win = EditorWindow(self.app, self.instance)
        win.attach(self)
        win.show_all()
        self.set_data("win", win)
        
    def show_in_main_window(self):
        self.instance.queries.attach(self)
        win = self.get_data("win")
        if win:
            win.destroy()
        self.set_data("win", None)
        
    def update_exectime(self, start, query):
        gobject.idle_add(self.lbl_status.set_text, "Query running... (%.3f seconds)" % (time.time()-start))
        if not query.executed:
            gobject.timeout_add(233, self.update_exectime, start, query)
        else:
            if query.failed:
                self.lbl_status.set_text(_(u"Query failed (%.3f seconds)") % query.execution_time)
            elif query.description:
                self.lbl_status.set_text(_(u"Query finished (%.3f seconds, %s rows)") % (query.execution_time, query.rowcount))
            else:
                self.lbl_status.set_text(_(u"Query finished (%.3f seconds, %s affected rows)") % (query.execution_time, query.rowcount))
            
        
    def update_textview_options(self):
        conf = self.app.config
        tv = self.textview
        buffer = tv.get_buffer()
        if USE_GTKSOURCEVIEW2:
            tv.set_show_right_margin(conf.get("editor.right_margin"))
            tv.set_right_margin_position(conf.get("editor.right_margin_position"))
            buffer.set_highlight_matching_brackets(conf.get("editor.bracket_matching"))
            tv.set_tab_width(conf.get("editor.tabs_width"))
            sm = gtksourceview2.style_scheme_manager_get_default()
            scheme = sm.get_scheme(conf.get("editor.scheme"))
            buffer.set_style_scheme(scheme)
        else:
            tv.set_show_margin(conf.get("editor.right_margin"))
            #tv.set_right_margin(conf.get("editor.right_margin_position"))
            buffer.set_check_brackets(conf.get("editor.bracket_matching"))
            tv.set_tabs_width(conf.get("editor.tabs_width"))
        if conf.get("editor.wrap_text"):
            if conf.get("editor.wrap_split"):
                tv.set_wrap_mode(gtk.WRAP_CHAR)
            else:
                tv.set_wrap_mode(gtk.WRAP_WORD)
        else:
            tv.set_wrap_mode(gtk.WRAP_NONE)
        tv.set_show_line_numbers(conf.get("editor.display_line_numbers"))
        tv.set_highlight_current_line(conf.get("editor.highlight_current_line"))
        tv.set_insert_spaces_instead_of_tabs(conf.get("editor.insert_spaces"))
        tv.set_auto_indent(conf.get("editor.auto_indent"))
        if conf.get("editor.default_font"):
            client = gconf.client_get_default()
            font = client.get_string("/desktop/gnome/interface/monospace_font_name")
        else:
            font = conf.get("editor.font")
        tv.modify_font(pango.FontDescription(font))
        
class EditorWindow(GladeWidget):
    
    def __init__(self, app, instance):
        GladeWidget.__init__(self, app, "crunchyfrog", "editorwindow")
        self.instance = instance
        self.editor = None
        self.toolbar = CFToolbar(self.app, "crunchyfrog", cb_provider=self)
        box = self.xml.get_widget("mainbox_editor")
        box.pack_start(self.toolbar.widget, False, False)
        box.reorder_child(self.toolbar.widget, 1)
        quit = self.toolbar.xml.get_widget("tb_quit")
        self.toolbar.widget.remove(quit)
        item = gtk.ToolButton()
        item.set_label(_(u"Show in main window"))
        item.set_stock_id("gtk-dnd-multiple")
        item.connect("clicked", self.on_show_in_main_window)
        item.show()
        self.toolbar.insert(item, -1)
        item = gtk.ToolButton()
        item.set_stock_id("gtk-close")
        item.connect("clicked", self.on_close)
        item.show()
        self.toolbar.insert(item, -1)
        self.toolbar.set_icon_size(gtk.ICON_SIZE_MENU)
        self.restore_window_state()
        
    def on_buffer_changed(self, buffer):
        gobject.idle_add(self.update_title)
        
    def on_close(self, *args):
        self.close()
        
    def on_query_new(self, *args):
        self.instance.on_query_new(self, *args)
        
    def on_copy(self, *args):
        self.editor.textview.get_buffer().copy_clipboard(gtk.clipboard_get())
        
    def on_paste(self, *args): 
        self.editor.textview.get_buffer().paste_clipboard(gtk.clipboard_get(), None, True)
        
    def on_cut(self, *args):
        self.editor.textview.get_buffer().cut_clipboard(gtk.clipboard_get(), True)
        
    def on_delete(self, *args):
        self.editor.textview.get_buffer().delete_selection(True, True)
        
    def on_show_in_main_window(self, *args):
        self.editor.show_in_main_window()
        
    def on_configure_event(self, win, event):
        config = self.app.config
        if not config.get("querywin.maximized"):
            config.set("querywin.width", event.width)
            config.set("querywin.height", event.height)
            
    def on_window_state_event(self, win, event):
        config = self.app.config
        bit = gtk.gdk.WINDOW_STATE_MAXIMIZED.value_names[0] in event.new_window_state.value_names
        config.set("querywin.maximized", bit)
        
    def attach(self, editor):
        self.editor = editor
        box = self.xml.get_widget("mainbox_editor")
        ebox = self.xml.get_widget("box_editor")
        ebox.get_parent().remove(ebox)
        if self.editor.get_parent():
            self.editor.reparent(box)
            expand, fill, padding, pack_type = box.query_child_packing(self.editor.widget)
            box.set_child_packing(self.editor.widget, True, True, padding, gtk.PACK_START)
        else:
            box.pack_start(self.editor.widget, True, True)
        box.reorder_child(self.editor.widget, 2)
        buffer = self.editor.textview.get_buffer()
        buffer.connect("changed", self.on_buffer_changed)
        self.toolbar.set_editor(editor)
        self.update_title()
        
    def close(self):
        self.destroy()
        
    def restore_window_state(self):
        if self.app.config.get("querywin.width", -1) != -1:
            self.widget.resize(self.app.config.get("querywin.width"),
                               self.app.config.get("querywin.height"))
        if self.app.config.get("querywin.maximized", False):
            self.widget.maximize()
        
    def update_title(self):
        buffer = self.editor.textview.get_buffer()
        txt = buffer.get_text(*buffer.get_bounds())
        txt = re.sub("\s+", " ", txt)
        txt = txt.strip()
        if len(txt) > 28:
            txt = txt[:25]+"..."
        if txt:
            txt = _(u"Query") + ": "+txt
        else:
            txt = _(u"Query")
        self.set_title(txt)
        

class ResultsView(GladeWidget):
    
    def __init__(self, app, instance, xml):
        GladeWidget.__init__(self, app, xml, "editor_results")
        self.instance = instance
        self.grid = ResultsGrid(app, instance, xml)
        self.messages = self.xml.get_widget("editor_results_messages")
        buffer = self.messages.get_buffer()
        buffer.create_tag("error", foreground="#a40000", weight=pango.WEIGHT_BOLD)
        
    def reset(self):
        buffer = self.messages.get_buffer()
        buffer.set_text("")
        self.set_current_page(2)
    
    def set_query(self, query):
        self.grid.set_query(query)
        buffer = self.messages.get_buffer()
        for err in query.errors:
            iter = buffer.get_end_iter()
            buffer.insert_with_tags_by_name(iter, err.strip()+"\n", "error")
        if query.errors:
            curr_page = 2
        elif query.description:
            curr_page = 0
        else:
            curr_page = 2
        gobject.idle_add(self.set_current_page, curr_page)
        
    def add_message(self, msg):
        buffer = self.messages.get_buffer()
        buffer.insert_at_cursor(msg.strip()+"\n")
        
        
class ResultsGrid(GladeWidget):
    
    def __init__(self, app, instance, xml):
        GladeWidget.__init__(self, app, xml, "editor_results_data")
        self.instance = instance
        self._idx = 0
        self._query = None
        
    def _setup_widget(self):
        self.grid = self.xml.get_widget("editor_resultsgrid")
        sel = self.grid.get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        self.grid.set_rubber_banding(True)
        
    def on_select_all_rows(self, *args):
        sel = self.grid.get_selection()
        sel.select_all()
        
    def on_show_more(self, *args):
        gobject.idle_add(self.fetch_next)
        
    def on_show_all(self, *args):
        gobject.idle_add(self.fetch_all)
        
    def set_query(self, query):
        self._query = query
        self._idx = 0
        while self.grid.get_columns():
            col = self.grid.get_column(0)
            self.grid.remove_column(col)
        btn_more = self.xml.get_widget("editor_results_more")
        btn_all = self.xml.get_widget("editor_results_all")
        btn_more.set_sensitive(False)
        btn_all.set_sensitive(False)
        btn_more.set_tooltip(self.instance.tt, _(u"Show next %(num)s rows") % {"num" : self.app.config.get("editor.results.offset")})
        btn_all.set_tooltip(self.instance.tt, _(u"Show all %(num)s rows") % {"num" : query.rowcount})
        if not query.description: return
        model_args = list()
        for name, type_code, display_size, internal_size, precision, scale, null_ok in query.description:
            model_args.append(str)
            renderer = gtk.CellRendererText()
            renderer.set_property("ellipsize", pango.ELLIPSIZE_END)
            renderer.set_property("wrap-width", 75)
            renderer.set_property("wrap-mode", gtk.WRAP_CHAR)
            renderer.set_property("single-paragraph-mode", True)
            renderer.set_property("width-chars", 10)
            col = gtk.TreeViewColumn(name.replace("_", "__"), renderer, markup=len(model_args)-1)
            col.set_sizing(gtk.TREE_VIEW_COLUMN_AUTOSIZE)
            col.set_resizable(True)
            self.grid.append_column(col)
        model_args.append(int)
        model_args.append(str)
        col = gtk.TreeViewColumn("#", gtk.CellRendererText(), text=len(model_args)-2, cell_background=len(model_args)-1)
        col.set_clickable(True)
        col.connect("clicked", self.on_select_all_rows)
        self.grid.insert_column(col, 0)
        model = gtk.ListStore(*model_args)
        self.grid.set_model(model)
        self.fetch_next()
        
    def fetch_all(self):
        while self.fetch_next():
            pass
        
    def fetch_next(self, fetch_all=False):
        model = self.grid.get_model()
        offset = self.app.config.get("editor.results.offset")
        style = self.grid.get_style()
        for i in range(self._idx, self._idx+offset):
            try:
                row = []
                for j in range(len(self._query.rows[i])):
                    value = self._query.rows[i][j]
                    if value == None: 
                        value = '<span foreground="%s">&lt;NULL&gt;</span>' % style.dark[gtk.STATE_PRELIGHT].to_string()
                    else:
                        value = gobject.markup_escape_text(str(value))
                    row.append(value)
                row.append(i+1)
                row.append(style.dark[gtk.STATE_ACTIVE].to_string())
                model.append(row)
            except IndexError:
                pass
        self._idx = self._idx+offset
        if self._idx > len(self._query.rows)-1:
            self._idx = len(self._query.rows)-1
        btn_more = self.xml.get_widget("editor_results_more")
        btn_all = self.xml.get_widget("editor_results_all")
        btn_more.set_sensitive(self._idx < len(self._query.rows)-1)
        btn_all.set_sensitive(self._idx < len(self._query.rows)-1)
        if fetch_all and self._idx < len(self._query.rows)-1:
            gobject.idle_add(self.fetch_next, fetch_all)
        else: 
            return (self._idx < len(self._query.rows)-1)
            



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: datasources.py 4 2008-02-26 05:04:22Z albrecht.andi $

"""Datasources"""

import gtk
import gobject

from kiwi.ui import dialogs

from gettext import gettext as _

from cf.ui import GladeWidget
from cf.datasources import DatasourceInfo

COMMON_OPTIONS = ["dsn", "database", "host", "port", "user", "password"]

class DatasourceManager(GladeWidget):
    
    def __init__(self, app):
        GladeWidget.__init__(self, app, "crunchyfrog", "datasourcemanager")
        if not self.app.plugins.get_plugins("crunchyfrog.backend", True):
            self.run = self.run_warning
        self.set_data("be_widgets", dict())
        
    def _setup_widget(self):
        # Fix help button
        btn = self.xml.get_widget("btn_help")
        box = self.xml.get_widget("dialog-action_area8")
        box.set_child_secondary(btn, True)
        # Backends
        cmb = self.xml.get_widget("cmb_backends")
        model = gtk.ListStore(gobject.TYPE_PYOBJECT,
                              str)
        model.set_sort_column_id(1, gtk.SORT_ASCENDING)
        cmb.set_model(model)
        cell = gtk.CellRendererText()
        cmb.pack_start(cell, True)
        cmb.add_attribute(cell, "text", 1)
        self._init_backends(model)
        # Saved connections
        tv = self.xml.get_widget("tv_stored_connections")
        model = gtk.ListStore(gobject.TYPE_PYOBJECT,
                              str)
        tv.set_model(model)
        col = gtk.TreeViewColumn("", gtk.CellRendererText(), markup=1)
        tv.append_column(col)
        sel = tv.get_selection()
        sel.connect("changed", self.on_selected_datasource_changed)
        self.refresh_saved_connections()
        
    def run_warning(self):
        dialogs.warning(_(u"No active database backends."),
                        _(u"Open preferences and activate at least one database backend plugin."))
        
    def _init_backends(self, model):
        model.clear()
        nb = self.xml.get_widget("nb_options")
        for be in self.app.plugins.get_plugins("crunchyfrog.backend", True):
            iter = model.append(None)
            model.set(iter, 0, be, 1, be.name)
            
    def _on_ask_for_password(self, check):
        return check.get_active()
            
    def _on_get_data_from_entry(self, entry):
        return entry.get_text().strip() or None
    
    def _on_get_port(self, spin):
        return spin.get_value_as_int() or None
            
    def _create_widget_database(self, data_widgets, conn=None):
        e = gtk.Entry()
        if conn and conn.options.get("database", None):
            e.set_text(conn.options.get("database"))
        data_widgets["database"] = (self._on_get_data_from_entry, e)
        return data_widgets, gtk.Label(_(u"Database:")), e
    
    def _create_widget_dsn(self, data_widgets, conn=None):
        e = gtk.Entry()
        if conn and conn.options.get("dsn", None):
            e.set_text(conn.options.get("dsn"))
        data_widgets["dsn"] = (self._on_get_data_from_entry, e)
        return data_widgets, gtk.Label(_(u"DSN:")), e
    
    def _create_widget_host(self, data_widgets, conn=None):
        e = gtk.Entry()
        if conn and conn.options.get("host", None):
            e.set_text(conn.options.get("host"))
        data_widgets["host"] = (self._on_get_data_from_entry, e)
        return data_widgets, gtk.Label(_(u"Host:")), e
    
    def _create_widget_port(self, data_widgets, conn=None):
        s = gtk.SpinButton(climb_rate=1, digits=0)
        s.set_range(0, 999999)
        s.set_increments(1, 10)
        if conn and conn.options.get("port", None):
            s.set_value(conn.options.get("port"))
        #self.app.ui.tt.set_tip(s, _(u"Setting port number to 0 means no port."))
        data_widgets["port"] = (self._on_get_port, s)
        return data_widgets, gtk.Label(_(u"Port:")), s
    
    def _create_widget_user(self, data_widgets, conn=None):
        e = gtk.Entry()
        if conn and conn.options.get("user", None):
            e.set_text(conn.options.get("user"))
        data_widgets["user"] = (self._on_get_data_from_entry, e)
        return data_widgets, gtk.Label(_(u"User:")), e
    
    def _create_widget_password(self, data_widgets, conn=None):
        def check_toggled(check, entry):
            entry.set_sensitive(not check.get_active())
        e = gtk.Entry()
        e.set_visibility(False)
        if conn and conn.options.get("password", None):
            e.set_text(conn.options.get("password"))
        data_widgets["password"] = (self._on_get_data_from_entry, e)
        check = gtk.CheckButton(_(u"_Ask for password"))
        check.connect("toggled", check_toggled, e)
        if conn:
            check.set_active(conn.options.get("ask_for_password", False))
        data_widgets["ask_for_password"] = (self._on_ask_for_password, check)
        box = gtk.VBox()
        box.pack_start(e, False)
        box.pack_start(check, False)
        return data_widgets, gtk.Label(_(u"Password:")), box
            
    def on_be_test_connection(self, btn):
        data = self.get_backend_options()
        lbl = self.xml.get_widget("lbl_testconnection")
        lbl.set_text("")
        combo = self.xml.get_widget("cmb_backends")
        iter = combo.get_active_iter()
        model = combo.get_model()
        be = model.get_value(iter, 0)
        err = be.test_connection(data)
        if err:
            dialogs.error(_(u"Connection failed"), err)
        else:
            lbl.set_text(_(u"Successful."))
            
    def on_cmb_backends_changed(self, combo):
        self.set_backend_option_widgets()
        
    def on_delete_datasource(self, *args):
        conn = self.get_selected_saved_connection()
        if not conn: return
        conn.delete()
        self.refresh_saved_connections()
            
    def on_new_datasource(self, *args):
        sel = self.xml.get_widget("tv_stored_connections").get_selection()
        sel.unselect_all()
        self.clear_fields()
        
    def on_save_datasource(self, *args):
        conn = self.get_connection()
        conn_info = self.get_selected_saved_connection()
        if conn_info:
            conn.db_id = conn_info.db_id
        conn.save()
        self.refresh_saved_connections(conn.db_id)
        
    def on_selected_datasource_changed(self, *args):
        conn = self.get_selected_saved_connection()
        self.xml.get_widget("btn_delete").set_sensitive(bool(conn))
        self.clear_fields()
        if not conn:
            return
        self.xml.get_widget("entry_name").set_text(conn.name or "")
        self.xml.get_widget("entry_description").set_text(conn.description or "")
        self.select_backend_by_id(conn.backend.id)
        self.set_backend_option_widgets(conn)
            
    def on_toggle_save_button(self, *args):
        btn = self.xml.get_widget("btn_save")
        cmb = self.xml.get_widget("cmb_backends")
        if self.xml.get_widget("entry_name").get_text().strip() \
        and self.xml.get_widget("cmb_backends").get_active_iter():
            btn.set_sensitive(True)
        else:
            btn.set_sensitive(False)
            
    def clear_fields(self):
        self.xml.get_widget("entry_name").set_text("")
        self.xml.get_widget("entry_description").set_text("")
        cmb = self.xml.get_widget("cmb_backends")
        cmb.set_active(-1)
        self.xml.get_widget("lbl_testconnection").set_text("")
            
    def get_backend_options(self):
        data_widgets = self.get_data("be_widgets")
        data = dict()
        for key, value in data_widgets.items():
            data[key] = value[0](value[1])
        return data
            
    def get_connection(self):
        conn = DatasourceInfo(self.app, self.get_selected_backend(),
                          self.xml.get_widget("entry_name").get_text().strip() or None,
                          self.xml.get_widget("entry_description").get_text().strip() or None,
                          self.get_backend_options())
        return conn
    
    def get_selected_backend(self):
        combo = self.xml.get_widget("cmb_backends")
        iter = combo.get_active_iter()
        if iter:
            model = combo.get_model()
            return model.get_value(iter, 0)
        else:
            return None
        
    def get_selected_saved_connection(self):
        sel = self.xml.get_widget("tv_stored_connections").get_selection()
        model, iter = sel.get_selected()
        if iter:
            return model.get_value(iter, 0)
        else:
            return None
        
    def refresh_saved_connections(self, active=None):
        model = self.xml.get_widget("tv_stored_connections").get_model()
        model.clear()
        for item in DatasourceInfo.load_all(self.app):
            iter = model.append(None)
            lbl = item.get_label()
            if item.description and item.description.strip():
                lbl += '\n<span size="small">'+item.description+'</span>'
            model.set(iter,
                      0, item,
                      1, lbl)
            if active and item.db_id == active:
                sel = self.xml.get_widget("tv_stored_connections").get_selection()
                sel.select_iter(iter)
            
    def select_backend_by_id(self, be_id):
        combo = self.xml.get_widget("cmb_backends")
        model = combo.get_model()
        iter = model.get_iter_first()
        while iter:
            be = model.get_value(iter, 0)
            if be.id == be_id:
                combo.set_active_iter(iter)
                return True
            iter = model.iter_next(iter)
        return False
    
    def set_backend_option_widgets(self, initial_data=None):
        be = self.get_selected_backend()
        data_widgets = dict()
        if be:
            data_widgets, widgets = be.get_datasource_options_widgets(data_widgets, initial_data)
        else:
            widgets = []
        vbox = self.xml.get_widget("vbox_be_options")
        while vbox.get_children():
            vbox.remove(vbox.get_children()[0])
        if not widgets:
            vbox.pack_start(gtk.Label(""))
        else:
            for widget in widgets:
                if widget in COMMON_OPTIONS:
                    data_widgets, lbl, x = getattr(self, "_create_widget_%s" % widget)(data_widgets, initial_data)
                    a = gtk.Alignment(0, 0)
                    a.add(lbl)
                    vbox.pack_start(a, False, False)
                    vbox.pack_start(x, False, False)
                else:
                    vbox.pack_start(widget, False, False)
        vbox.show_all()
        self.set_data("be_widgets", data_widgets)
        self.xml.get_widget("btn_test_connection").set_sensitive(bool(be))


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: datasources.py 16 2008-02-26 14:52:30Z albrecht.andi $

import gobject
import cPickle

import logging
log = logging.getLogger("DS")

class DatasourceInfo(gobject.GObject):
    
    def __init__(self, app, backend, name=None, description=None, options={}):
        self.__gobject_init__()
        self.app = app
        self.db_id = None
        self.backend = backend
        self.name = name
        self.description = description
        self.options = options
        self.__conncount = 1L
        self.__connections = list()
        self.internal_connection = None
        
    def __cmp__(self, other):
        return cmp(self.db_id, other.db_id)
        
    def get_label(self):
        return self.backend.get_label(self)
    
    def save(self):
        cur = self.app.userdb.cursor
        conn = self.app.userdb.conn
        if not self.db_id:
            sql = "insert into datasource (name, description, backend, options) \
            values (?,?,?,?)"
            cur.execute(sql, (self.name, self.description, self.backend.id,
                              self.serialize_options()))
            self.db_id = cur.lastrowid
            conn.commit()
            self.app.datasources.emit("datasource-added", self)
        else:
            sql = "update datasource set name=?, description=?, backend=?, \
            options=? where id=?"
            cur.execute(sql, (self.name, self.description, self.backend.id,
                              self.serialize_options(), self.db_id))
            conn.commit()
            self.app.datasources.emit("datasource-modified", self)
            
    def delete(self):
        cur = self.app.userdb.cursor
        conn = self.app.userdb.conn
        sql = "delete from datasource where id=?"
        cur.execute(sql, (self.db_id,))
        conn.commit()
        self.app.datasources.emit("datasource-deleted", self)
            
    def serialize_options(self):
        return cPickle.dumps(self.options)
    
    def deserialize_options(self, data):
        return cPickle.loads(data)
    
    @classmethod
    def load(cls, app, db_id):
        cur = app.userdb.cursor
        sql = "select name, description, backend, options from datasource \
        where id=?"
        cur.execute(sql, (db_id,))
        res = cur.fetchone()
        backend = app.plugins.by_id(res[2])
        opts = cPickle.loads(str(res[3]))
        i = cls(app, backend, res[0], res[1], opts)
        i.db_id = db_id
        return i
    
    @classmethod
    def load_all(cls, app):
        cur = app.userdb.cursor
        sql = "select id from datasource"
        cur.execute(sql)
        r = list()
        for item in cur.fetchall():
            r.append(cls.load(app, item[0]))
        return r
    
    def get_connections(self):
        return self.__connections
    
    def add_connection(self, connection):
        connection.connect("closed", self.on_connection_closed)
        self.__connections.append(connection)
        self.app.datasources.emit("datasource-modified", self)
        
    def on_connection_closed(self, connection):
        if connection in self.__connections:
            self.__connections.remove(connection)
        gobject.idle_add(self.app.datasources.emit, "datasource-modified", self)
            
    def dbconnect(self):
        conn = self.backend.dbconnect(self.options)
        conn.datasource_info = self
        conn.conn_number = self.__conncount
        self.__conncount += 1
        if not self.internal_connection:
            self.internal_connection = conn
        self.add_connection(conn)
        return conn
    
    def dbdisconnect(self):
        while self.__connections:
            conn = self.__connections[0]
            log.info("Closing connection %s" % conn)
            conn.close()
    
class DatasourceManager(gobject.GObject):
    """Datasource manager
    
    An instance of the ``DatasourceManager`` class is available
    as the ``datasources`` instance attribute of `CFApplication`.
    
    Signals
    =======
        datasource-added
            ```def callback(manager, datasource_info, user_param1, ...)```
            
            Emitted when a datsource was added
            
        datasource-deleted
            ```def callback(manager, datasource_info, user_param1, ...)```
            
            Emitted when a datasource was removed
            
        datasource-modified
            ```def callback(manager, datasource_info, user_param1, ...)```
            
            Emitted when a datasource was modified
            
    """
    
    __gsignals__ = {
        "datasource-added" : (gobject.SIGNAL_RUN_LAST,
                              gobject.TYPE_NONE,
                              (gobject.TYPE_PYOBJECT,)),
        "datasource-deleted" : (gobject.SIGNAL_RUN_LAST,
                              gobject.TYPE_NONE,
                              (gobject.TYPE_PYOBJECT,)),
        "datasource-modified" : (gobject.SIGNAL_RUN_LAST,
                              gobject.TYPE_NONE,
                              (gobject.TYPE_PYOBJECT,)),
    }
    
    def __init__(self, app):
        self.app = app
        self._cache = list()
        self.__gobject_init__()
        self.connect("datasource-added", self.on_datasource_added)
        self.connect("datasource-deleted", self.on_datasource_deleted)
        self.connect("datasource-modified", self.on_datasource_modified)
        self.app.plugins.connect("plugin-active", self.on_plugin_active)
        for item in DatasourceInfo.load_all(self.app):
            if not item.backend or not self.app.plugins.is_active(item.backend):
                continue
            self._cache.append(item)
        
    def on_datasource_added(self, manager, datasource_info):
        self._cache.append(datasource_info)
    
    def on_datasource_deleted(self, manager, datasource_info):
        if datasource_info in self._cache:
            self._cache.remove(datasource_info)
    
    def on_datasource_modified(self, manager, datasource_info):
        for ds in self._cache:
            if ds != datasource_info:
                continue
            ds.backend = datasource_info.backend
            ds.description = datasource_info.description
            ds.name = datasource_info.name
            ds.options = datasource_info.options
            
    def on_plugin_active(self, manager, plugin, active):
        if not plugin._entry_point_group == "crunchyfrog.backend":
            return
        for item in DatasourceInfo.load_all(self.app):
            if item.backend and active and item not in self._cache:
                self.emit("datasource-added", item)
            elif not active and item in self._cache:
                item.dbdisconnect()
                self.emit("datasource-deleted", item)
        
    def get_all(self):
        return self._cache
        
    
def check_userdb(userdb):
    if not userdb.get_table_version("datasource"):
        sql = "create table datasource (id integer primary key, \
        name text, description text, backend text, options text, \
        last_accessed real, num_accessed integer)"
        userdb.create_table("datasource", "0.1", sql)


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: __init__.py 4 2008-02-26 05:04:22Z albrecht.andi $

import gobject
import gtk

from os.path import abspath, join, dirname

import cf

class GladeWidget(gobject.GObject):
    
    def __init__(self, app, xml, widget_name,
                 signal_autoconnect=True,
                 cb_provider=None):
        self.__gobject_init__()
        self.app = app
        if isinstance(xml, basestring):
            if not xml.endswith(".glade"):
                xml += ".glade"
            self.xml = gtk.glade.XML(join(cf.DATA_DIR, xml), widget_name)
        else:
            self.xml = xml
        if signal_autoconnect:
            if not cb_provider:
                cb_provider = self
            self.xml.signal_autoconnect(cb_provider)
        self.widget = self.xml.get_widget(widget_name)
        for item in dir(self.widget):
            if item.startswith("_") \
            or hasattr(self, item): continue
            setattr(self, item, getattr(self.widget, item))
        self.widget.set_data("glade-widget", self)
        self._setup_widget()
        self._setup_connections()
        
    def _setup_widget(self):
        """Called when the object is created."""
        
    def _setup_connections(self):
        """Called when the object is created."""
        


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: instance.py 16 2008-02-26 14:52:30Z albrecht.andi $

import gtk
import gtk.glade
import gnome
import gdl
import gobject

from gettext import gettext as _

from cf import release
from cf.ui import GladeWidget
from cf.ui import pdock
from cf.ui.browser import Browser
from cf.ui.datasources import DatasourceManager
from cf.ui.editor import Editor, EditorWindow
from cf.ui.queries import QueriesNotebook
from cf.ui.statusbar import CFStatusbar
from cf.ui.toolbar import CFToolbar
from cf.ui.widgets import ConnectionButton

class CFInstance(GladeWidget):
    
    def __init__(self, app):
        GladeWidget.__init__(self, app, "crunchyfrog", "mainwindow")
        self._editor = None
        
    def _setup_widget(self):
        # Window state
        if self.app.config.get("gui.width", -1) != -1:
            self.widget.resize(self.app.config.get("gui.width"),
                               self.app.config.get("gui.height"))
        if self.app.config.get("gui.maximized", False):
            self.widget.maximize()
        # Tooltips
        self.tt = gtk.Tooltips()
        # Toolbar
        self.toolbar = CFToolbar(self.app, self.xml)
        # Statusbar
        self.statusbar = CFStatusbar(self.app, self.xml)
        # Dock
        self.dock = pdock.Dock()
        box = self.xml.get_widget("box_main")
        box.pack_start(self.dock, True, True)
        box.reorder_child(self.dock, 2)
        # Queries
        self.queries = QueriesNotebook(self.app, self)
        item = pdock.DockItem(self.dock, "queries", self.queries, _(u"Queries"),
                              "gtk-edit", None, pdock.DOCK_ITEM_BEH_LOCKED)
        self.dock.add_item(item)
        # Browser
        self.browser = Browser(self.app, self)
        item = pdock.DockItem(self.dock, "browser", self.browser, _(u"Browser"), 
                              "gtk-find", gtk.POS_LEFT)
        self.dock.add_item(item)
    
    def _init_ui(self, argv):
        pass
    
    def on_about(self, *args):
        def open_url(dialog, url):
            gnome.url_show(url)
        gtk.about_dialog_set_url_hook(open_url)
        dlg = gtk.AboutDialog()
        dlg.set_name(release.name)
        dlg.set_version(release.version)
        dlg.set_copyright(release.copyright)
        dlg.set_license(release.license)
        dlg.set_website(release.url)
        dlg.set_website_label(release.url)
        dlg.set_logo_icon_name(release.appname)
        dlg.set_program_name(release.appname)
        dlg.run()
        dlg.destroy()
        
    def on_commit(self, *args):
        if not self._editor:
            return 
        gobject.idle_add(self._editor.commit)
        
    def on_rollback(self, *args):
        if not self._editor:
            return
        gobject.idle_add(self._editor.rollback)
        
    def on_begin_transaction(self, *args):
        if not self._editor:
            return
        gobject.idle_add(self._editor.begin_transaction)
        
    def on_configure_event(self, win, event):
        config = self.app.config
        if not config.get("gui.maximized"):
            config.set("gui.width", event.width)
            config.set("gui.height", event.height)
            
    def on_copy(self, *args):
        if not self._editor:
            return
        self._editor.textview.get_buffer().copy_clipboard(gtk.clipboard_get())
        
    def on_paste(self, *args):
        if not self._editor:
            return 
        self._editor.textview.get_buffer().paste_clipboard(gtk.clipboard_get(), None, True)
        
    def on_cut(self, *args):
        if not self._editor:
            return
        self._editor.textview.get_buffer().cut_clipboard(gtk.clipboard_get(), True)
        
    def on_delete(self, *args):
        if not self._editor:
            return
        self._editor.textview.get_buffer().delete_selection(True, True)
            
    def on_datasource_manager(self, *args):
        dlg = DatasourceManager(self.app)
        dlg.run()
        dlg.destroy()
        
    def on_execute_query(self, *args):
        gobject.idle_add(self._editor.execute_query)
        
    def on_help(self, *args):
        self.show_help()
    
    def on_new_instance(self, *args):
        self.app.new_instance(tuple())
        
    def on_report_problem(self, *args):
        gobject.idle_add(self.open_website, "http://code.google.com/p/crunchyfrog/issues/list")
        
    def on_open_devpages(self, *args):
        gobject.idle_add(self.open_website, "http://crunchyfrog.googlecode.com")
        
    def on_open_helptranslate(self, *args):
        gobject.idle_add(self.open_website, "https://translations.launchpad.net/crunchyfrog/trunk/")
        
    def on_preferences(self, *args):
        self.app.preferences_show()
        
    def on_query_new(self, *args):
        editor = Editor(self.app, self)
        if self.app.config.get("editor.open_in_window"):
            editor.show_in_separate_window()
        else:
            self.queries.attach(editor)
        editor.show_all()
        
    def on_quit(self, *args):
        self.widget.destroy()
        
    def on_window_state_event(self, win, event):
        config = self.app.config
        bit = gtk.gdk.WINDOW_STATE_MAXIMIZED.value_names[0] in event.new_window_state.value_names
        config.set("gui.maximized", bit)
        
    def open_website(self, url):
        gnome.url_show(url)
        
    def set_editor_active(self, editor, active):
        if not active:
            editor = None
        self._editor = editor
        self.toolbar.set_editor(editor)
        
    def show_help(self, topic=None):
        gnome.help_display(release.appname, topic)


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: __init__.py 4 2008-02-26 05:04:22Z albrecht.andi $

"""Configuration"""

import gobject

from os.path import abspath, dirname, join
from configobj import ConfigObj

from gettext import gettext as _

from cf import USER_CONF

class Config(gobject.GObject):
    """Configuration object"""
    
    __gsignals__ = {
        "changed" : (gobject.SIGNAL_RUN_LAST,
                     gobject.TYPE_NONE,
                     (str, gobject.TYPE_PYOBJECT)),
    }
    
    def __init__(self, app):
        self.app = app
        self.__gobject_init__() # IGNORE:E1101
        self.__conf = None
        self.__init_conf()
        self.app.register_shutdown_task(self.on_app_shutdown, 
                                        _(u"Writing configuration"))
        
    def on_app_shutdown(self, *args): # IGNORE:W0613
        """Callback: write configuration file to disk"""
        self.write()
        
    def __init_conf(self):
        """Intialize the configuration system"""
        self.__conf = ConfigObj(abspath(join(dirname(__file__), "default.cfg")),
                                unrepr=True)
        self.__conf.update(ConfigObj(USER_CONF, unrepr=True))
        
    def init(self):
        """Loads configuration"""
        pass
        
    def get(self, key, default=None):
        """Returns value or default for key"""
        return self.__conf.get(key, default)
    
    def set(self, key, value):
        """Sets key to value"""
        self.__conf[key] = value
        self.emit("changed", key, value) # IGNORE:E1101
        
    def write(self, fname=None):
        """Writes configuration file"""
        if not fname:
            fname = USER_CONF
        fp = open(USER_CONF, "w")
        self.__conf.write(fp)
        fp.close()


# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: __init__.py 4 2008-02-26 05:04:22Z albrecht.andi $

"""Shared widgets"""

import gtk
import gobject
import gdl
import pango

from gettext import gettext as _

class ConnectionButton(gdl.ComboButton):
    
    def __init__(self, app):
        gdl.ComboButton.__init__(self)
        self.app = app
        self._setup_widget()
        self.set_editor(None)
        self.app.datasources.connect("datasource-modified", self.on_datasource_modified)
        
    def _setup_widget(self):
        lbl = self.get_children()[0].get_children()[0].get_children()[0].get_children()[1]
        lbl.set_max_width_chars(25)
        lbl.set_ellipsize(pango.ELLIPSIZE_END)
        lbl.set_use_markup(True)
        self.label = lbl
        self._menu = gtk.Menu()
        self.set_menu(self._menu)
        
    def on_datasource_modified(self, *args):
        self.rebuild_menu()
        
    def on_new_connection(self, item, datasource_info):
        conn = datasource_info.dbconnect()
        self._editor.set_connection(conn)
        self.set_editor(self._editor)
        
    def on_set_connection(self, item, connection):
        self._editor.set_connection(connection)
        self.set_editor(self._editor)
        
    def rebuild_menu(self):
        while self._menu.get_children():
            self._menu.remove(self._menu.get_children()[0])
        if not self._editor:
            return
        ghas_connections = False
        for datasource_info in self.app.datasources.get_all():
            item = gtk.MenuItem(datasource_info.get_label())
            item.show()
            self._menu.append(item)
            menu = gtk.Menu()
            has_connections = False
            for conn in datasource_info.get_connections():
                yitem = gtk.MenuItem(_(u"Connection")+" #%s" % conn.conn_number)
                yitem.connect("activate", self.on_set_connection, conn)
                yitem.show()
                menu.append(yitem)
                has_connections = True
                ghas_connections = True
            if has_connections:
                sep = gtk.SeparatorMenuItem()
                sep.show()
                menu.append(sep)
            xitem = gtk.MenuItem(_(u"New connection"))
            xitem.connect("activate", self.on_new_connection, datasource_info)
            xitem.show()
            menu.append(xitem)
            item.set_submenu(menu)
        sep = gtk.SeparatorMenuItem()
        sep.show()
        self._menu.append(sep)
        item = gtk.MenuItem(_(u"Manage open connections"))
        item.set_sensitive(ghas_connections)
        item.show()
        self._menu.append(item)    
        
    def set_editor(self, editor):
        self._editor = editor
        self.set_sensitive(bool(editor))
        self.rebuild_menu()
        if editor:
            if editor.connection:
                self.set_label(editor.connection.get_label())
                markup = "<b>"+editor.connection.datasource_info.get_label()+"</b>\n"
                markup += _(u"Connection")+" #%s" % editor.connection.conn_number
                self.set_tooltip_markup(markup)
            else:
                self.set_label("<"+_(u"Not connected")+">")
                self.set_tooltip_markup(_(u"Click to open a connection"))
                 


from setuptools import setup, find_packages
import os

from cf import release

data_files = [
    ("share/pixmaps", [os.path.abspath("data/crunchyfrog.png")]),
    ("man/man1", ["data/crunchyfrog.1"]),
    ("share/applications", ["data/crunchyfrog.desktop"]),
    ("share/crunchyfrog", [os.path.abspath("data/crunchyfrog.glade")]),
    ("share/icons/hicolor/scalable/apps", [os.path.abspath("data/crunchyfrog.svg")])
]
for item in os.listdir("po/"):
    if not os.path.isdir(os.path.join("po/", item)) \
    or item == ".svn" \
    or not os.path.isfile(os.path.join("po/", item, "LC_MESSAGES/crunchyfrog.mo")):
        continue
    data_files.append(("share/locale/%s/LC_MESSAGES" % item, ["po/%s/LC_MESSAGES/crunchyfrog.mo" % item]))

setup(
    data_files=data_files,
    name=release.appname,
    version=release.version,
    description=release.description,
    author=release.author,
    author_email=release.author_email,
    long_description=release.long_description,
    url=release.url,
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Environment :: X11 Applications :: Gnome",
        "Environment :: X11 Applications :: GTK",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: Database :: Front-Ends",
        "Topic :: Desktop Environment :: Gnome",
    ],
    zip_safe=False,
    packages=find_packages(),
    include_package_data=True,
    install_requires = [
        "ConfigObj >= 4.4.0",
        "lxml >= 0.9", # pdock requirement
        "kiwi >= 1.9.16", 
    ],
    entry_points="""
    [gui_scripts]
    crunchyfrog = cf:main
    
    [crunchyfrog.backend]
    postgres = cf.backends.postgres:PostgresBackend
    sqlite = cf.backends.sqlite:SQLiteBackend
    mysql = cf.backends.mysql:MySQLBackend
    oracle = cf.backends.oracle:OracleBackend
    """
)



# -*- coding: utf-8 -*-

# crunchyfrog - a database schema browser and query tool
# Copyright (C) 2008 Andi Albrecht <albrecht.andi@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# $Id: core.py 4 2008-02-26 05:04:22Z albrecht.andi $

"""Core objects"""

import gobject
import gnomevfs
import pkg_resources
import sys
import os
from kiwi.ui import dialogs

import logging
log = logging.getLogger("PLUGINS")

from gettext import gettext as _

from cf import USER_PLUGIN_DIR, PLUGIN_DIR

class GenericPlugin(gobject.GObject):
    
    name = None
    description = None
    long_description = None
    icon = None
    author = None
    license = None
    homepage = None
    version = None
    
    def __init__(self, app):
        self.app = app
        self.__gobject_init__()
        
    def shutdown(self):
        pass
        
class DBBackendPlugin(GenericPlugin):
    icon = "stock_database"
    
    def __init__(self, app):
        GenericPlugin.__init__(self, app)
        self.schema = None
    
    @classmethod
    def get_datasource_options_widgets(cls, data_widgets, initial_data=None):
        return data_widgets, []
    # TODO: Documentation
    
    @classmethod
    def get_label(cls, datasource_info):
        s = "%s@%s on %s" % (datasource_info.options.get("user", None) or "??",
                             datasource_info.options.get("database", None) or "??",
                             datasource_info.options.get("host", None) or "??")
        if datasource_info.name:
            s = '%s (%s)' % (datasource_info.name, s)
        return s
    
    def test_connection(self, data):
        raise NotImplementedError
    
    def password_prompt(self):
        return dialogs.password(_(u"Password required"), _(u"Enter the password to connect to this database."))
        
# Available entry points
# Key: entry point name
# Value: 2-tuple (label, expected class)
ENTRY_POINTS = {
    "crunchyfrog.plugin" : (_(u"Miscellaneous"), GenericPlugin),
    "crunchyfrog.backend" : (_(u"Database backends"), DBBackendPlugin),
} 

class PluginManager(gobject.GObject):
    """Plugin manager"""
    
    entry_points = ENTRY_POINTS
    
    __gsignals__ = {
        "plugin-added" : (gobject.SIGNAL_RUN_LAST,
                          gobject.TYPE_NONE,
                          (gobject.TYPE_PYOBJECT,)),
        "plugin-removed" : (gobject.SIGNAL_RUN_LAST,
                            gobject.TYPE_NONE,
                            (gobject.TYPE_PYOBJECT,)),
        "plugin-active" : (gobject.SIGNAL_RUN_LAST,
                           gobject.TYPE_NONE,
                           (gobject.TYPE_PYOBJECT, bool)),
    }
    
    def __init__(self, app):
        self.app = app
        self.__gobject_init__()
        self.__plugins = dict()
        self.__active_plugins = dict()
        gnomevfs.monitor_add(USER_PLUGIN_DIR, gnomevfs.MONITOR_DIRECTORY, self.on_plugin_folder_changed)
        gnomevfs.monitor_add(PLUGIN_DIR, gnomevfs.MONITOR_DIRECTORY, self.on_plugin_folder_changed)
        self.refresh()
        
    def on_plugin_folder_changed(self, folder, path, change):
        if change in [gnomevfs.MONITOR_EVENT_DELETED, gnomevfs.MONITOR_EVENT_CREATED]:
            gobject.idle_add(self.refresh)
        
    def get_plugins(self, entry_point_name, active_only=False):
        ret = list()
        if active_only:
            plugins = self.__active_plugins.values()
        else:
            plugins = self.__plugins.values()
        for item in plugins:
            if item._entry_point_group == entry_point_name:
                ret.append(item)
        return ret
    
    def refresh(self):
        # cleanup sys.path to remove deleted eggs
        for path in sys.path:
            if path.endswith(".egg") and not os.path.isfile(path):
                sys.path.remove(path)
        working_set = pkg_resources.WorkingSet()
        working_set.add_entry(PLUGIN_DIR)
        working_set.add_entry(USER_PLUGIN_DIR)
        env = pkg_resources.Environment([PLUGIN_DIR, USER_PLUGIN_DIR])
        dists, errors = working_set.find_plugins(env)
        # TODO: Handle errors
        map(working_set.add, dists)
        for path in working_set.entries:
            if path not in sys.path:
                sys.path.append(path)
        dist = pkg_resources.get_distribution("crunchyfrog")
        ids_found = list()
        for entry_point_name in self.entry_points.keys():
            for entry_point in working_set.iter_entry_points(entry_point_name):
                id = ".".join([entry_point_name, entry_point.name])
                try:
                    plugin = entry_point.load()
                except:
                    import traceback; traceback.print_exc()
                    continue
                if not self.__plugins.has_key(id):
                    added = True
                else:
                    added = False
                plugin._entry_point = entry_point
                plugin._entry_point_group = entry_point_name
                plugin.id = ".".join([entry_point_name, entry_point.name])
                self.__plugins[id] = plugin
                ids_found.append(id)
                if added:
                    l = self.app.config.get("plugins.active", [])
                    if id in l:
                        self.set_active(plugin, True)
                    self.emit("plugin-added", plugin)
        for module_name, plugin in self.__plugins.items():
            entry_point = plugin._entry_point
            group = plugin._entry_point_group
            id = ".".join([group, entry_point.name])
            if id not in ids_found:
                l = self.app.config.get("plugins.active", [])
                if id in l:
                    self.set_active(plugin, False)
                del self.__plugins[id]
                self.emit("plugin-removed", plugin)
    
    def set_active(self, plugin, active):
        id = None
        for key, value in self.__plugins.items():
            if value == plugin:
                id = key
                break
        if not id: 
            return
        l = self.app.config.get("plugins.active", [])
        if active:
            x = plugin(self.app)
            self.__active_plugins[plugin] = x
            if id not in l:
                l.append(id)
        elif not active:
            x = self.__active_plugins.get(plugin, None)
            if x:
                x.shutdown()
                del self.__active_plugins[plugin]
            if id in l:
                l.remove(id)
        self.app.config.set("plugins.active", l)
        self.emit("plugin-active", plugin, active)
        
    def is_active(self, plugin):
        if isinstance(plugin, GenericPlugin):
            plugin = plugin.__class__
        return self.__active_plugins.has_key(plugin)
    
    def by_id(self, id, active_only=True):
        if active_only:
            plugins = self.__active_plugins.values()
        else:
            plugins = self.__plugins.values()
        for plugin in plugins:
            if plugin.id == id:
                return plugin
        return None


from setuptools import setup

setup(
    name="CrunchyFrog example plugin",
    version="0.1",
    py_modules=["example"],
    entry_points="""
    [crunchyfrog.plugin]
    foo = example:ExamplePlugin
    """
)